from dataclasses import dataclass
from typing import ClassVar, Generic, Type, TypeVar, cast, get_args, get_origin
import json
import base64
import uuid

import caseconverter
import falcon
import jsonpatch

from kuroboros import logger as klogger
from kuroboros.exceptions import MutationWebhookError, ValidationWebhookError
from kuroboros.extended_api import ExtendedClient
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.schema import CRDSchema

T = TypeVar("T", bound=CRDSchema)


class WebhookTypes:
    """
    The available webhook types
    """

    VALIDATION = "Validation"
    MUTATION = "Mutation"


class OperationsEnum:
    """
    Enum of kubernetes operations
    """

    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"


@dataclass
class Request(Generic[T]):
    """
    AdmissionReview v1 request.
    https://kubernetes.io/docs/reference/access-authn-authz/extensible-admission-controllers/#request
    """

    uid: str
    kind: dict
    resource: dict
    sub_resource: str | None
    request_kind: dict
    request_resource: dict
    request_sub_resource: str | None
    name: str
    namespace: str | None
    user_info: dict
    object: T | None
    old_object: T | None
    options: dict | None
    dry_run: bool
    operation: str


class BaseWebhook(Generic[T]):
    """
    The Base Webhook Class, all webhook should implement this
    """

    __group_version_info: ClassVar[GroupVersionInfo]
    _generic_base_type: ClassVar = None

    name: str
    logger = klogger.root_logger.getChild(__name__)
    register_on = []

    _endpoint: str
    _type: Type[T]
    _webhook_type: str
    _endpoint_suffix: str
    _extended_client: ExtendedClient

    @classmethod
    def set_gvi(cls, gvi: GroupVersionInfo) -> None:
        """
        Sets the GroupVersionInfo of the Reconciler
        """
        cls.__group_version_info = gvi

    @classmethod
    def get_config_dict(cls):
        """
        Returns the config dict to generate the kubernetes YAML
        """
        return {
            "crd_name": cls.__group_version_info.crd_name,
            "group": cls.__group_version_info.group,
            "singular": cls.__group_version_info.singular,
            "plural": cls.__group_version_info.plural,
            "api_version": cls.__group_version_info.api_version,
            "operations": cls.register_on,
            "scope": cls.__group_version_info.scope,
        }

    @classmethod
    def crd_type(cls) -> Type[T]:
        """
        Returns the type of CRD that this webhook handle
        """
        t_type = None
        for base in getattr(cls, "__orig_bases__", []):
            origin = get_origin(base)
            if origin in (BaseValidationWebhook, BaseMutationWebhook):
                t_type = get_args(base)[0]
                break

        if t_type is None or CRDSchema not in t_type.__mro__:
            raise RuntimeError("Could not determine generic type T. ")

        return t_type

    @property
    def endpoint(self) -> str:
        """
        Returns the endpoint to this webhook.
        Its formed by <apiVersion>/<singular>/<webhook_suffix>
        """
        gvi = self.__group_version_info
        return f"/{gvi.api_version}/{gvi.singular}/{self._endpoint_suffix}"

    def __init__(self) -> None:
        pretty_version = self.__group_version_info.pversion()
        singular = self.__group_version_info.singular.capitalize()
        self.name = caseconverter.pascalcase(
            f"{singular}{pretty_version}{self._webhook_type}Webhook"
        )
        self.logger = self.logger.getChild(self.name)
        self._extended_client = ExtendedClient()

    def process(self, body: bytes):
        """
        Processess the raw request
        """
        raise NotImplementedError("Subclasses must implement the process method")

    def on_post(self, req: falcon.Request, resp: falcon.Response) -> None:
        """
        POST on `endpoint`
        """
        raw = req.stream.read()
        uid = json.loads(raw.decode("utf-8")).get("request", {}).get("uid", str(uuid.uuid4()))
        try:
            response, status, headers = self.process(raw)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(e)
            response, status, headers = (
                json.dumps(
                    {
                        "apiVersion": "admission.k8s.io/v1",
                        "kind": "AdmissionReview",
                        "response": {
                            "uid": uid,
                            "allowed": False,
                            "status": {
                                "message": f"{self._webhook_type} webhook error"
                            },
                        },
                    }
                ),
                falcon.HTTP_500,
                {"Content-Type": "application/json"},
            )
        resp.status = status
        resp.text = response
        for k, v in (headers or {}).items():
            resp.set_header(k, v)

    def _serialize(self, obj):
        return self._extended_client.sanitize_for_serialization(obj)


class BaseValidationWebhook(BaseWebhook, Generic[T]):
    """
    Kuroboros BaseValidationWebhook.
    Registers an endpoint in /<apiVersion>/<singular>/validate
    """

    _webhook_type = WebhookTypes.VALIDATION
    _endpoint_suffix = "validate"

    register_on = [OperationsEnum.CREATE, OperationsEnum.UPDATE, OperationsEnum.DELETE]

    def validate(self, request: Request[T]) -> None:
        """
        Define validation logic
        """

    def process(self, body: bytes):
        self.logger.debug("processing validation webhook")
        request: Request[T] | None = None
        admission_review = json.loads(body.decode("utf-8"))
        r = admission_review.get("request", {})
        try:
            obj = self._extended_client.deserialize(r.get("object"), self.crd_type())
            old_obj = self._extended_client.deserialize(
                r.get("oldObject"), self.crd_type()
            )
            operation = r.get("operation")

            if operation not in self.register_on:
                raise ValidationWebhookError(f"unsupported operation: {operation}")

            request = Request(
                uid=r["uid"],
                name=r["name"],
                namespace=r.get("namespace"),
                kind=r["kind"],
                request_kind=r["requestKind"],
                request_resource=r["requestResource"],
                request_sub_resource=r.get("requestSubResource"),
                resource=r["resource"],
                sub_resource=r.get("subResource"),
                user_info=r["userInfo"],
                options=r.get("options"),
                dry_run=r["dryRun"],
                object=cast(T, obj),
                old_object=cast(T, old_obj),
                operation=operation,
            )

            self.validate(request)

            self.logger.debug("validation passed")
            response = {
                "apiVersion": "admission.k8s.io/v1",
                "kind": "AdmissionReview",
                "response": {"uid": r.get("uid"), "allowed": True},
            }
            return (
                json.dumps(response),
                falcon.HTTP_200,
                {"Content-Type": "application/json"},
            )
        except ValidationWebhookError as e:
            self.logger.warning(f"validation failed: {e.reason}")
            response = {
                "apiVersion": "admission.k8s.io/v1",
                "kind": "AdmissionReview",
                "response": {
                    "uid": r.get("uid") if request else None,
                    "allowed": False,
                    "status": {"message": e.reason},
                },
            }
            return (
                json.dumps(response),
                falcon.HTTP_200,
                {"Content-Type": "application/json"},
            )


class BaseMutationWebhook(BaseWebhook, Generic[T]):
    """
    Kuroboros BaseMutationWebhook.
    Registers an endpoint in /<apiVersion>/<singular>/mutate
    """

    _webhook_type = WebhookTypes.MUTATION
    _endpoint_suffix = "mutate"

    register_on = [OperationsEnum.CREATE, OperationsEnum.UPDATE]

    def __init__(self):
        for reg in self.register_on:
            if reg not in (OperationsEnum.CREATE, OperationsEnum.UPDATE):
                raise RuntimeError(
                    "unsupported type to register.",
                    "supported `CREATE` or `UPDATE`",
                    reg,
                )
        super().__init__()

    def mutate(self, request: Request[T]) -> T:
        """
        Define the mutation logic
        """
        assert request.object is not None
        return request.object

    def process(self, body: bytes):
        self.logger.debug("processing mutation webhook")
        request: Request[T] | None = None
        admission_review = json.loads(body.decode("utf-8"))
        r = admission_review.get("request", {})
        try:
            obj = cast(
                T, self._extended_client.deserialize(r.get("object"), self.crd_type())
            )
            old_obj = cast(
                T | None,
                self._extended_client.deserialize(r.get("oldObject"), self.crd_type()),
            )
            operation = r.get("operation")

            crd_instance = self.crd_type()(**obj.to_dict())

            if operation not in self.register_on:
                raise MutationWebhookError(f"unsupported operation: {operation}")

            request = Request(
                uid=r["uid"],
                name=r["name"],
                namespace=r.get("namespace"),
                kind=r["kind"],
                request_kind=r["requestKind"],
                request_resource=r["requestResource"],
                request_sub_resource=r.get("requestSubResource"),
                resource=r["resource"],
                sub_resource=r.get("subResource"),
                user_info=r.get("userInfo"),
                options=r.get("options"),
                dry_run=r.get("dryun"),
                object=self.crd_type()(**obj.to_dict()),
                old_object=(self.crd_type()(**old_obj.to_dict()) if old_obj else None),
                operation=operation,
            )

            mutated_crd = self.mutate(request)
            assert mutated_crd is not None, "Mutated CRD instance cannot be None"
            patch_ops = jsonpatch.JsonPatch.from_diff(
                self._extended_client.serialize(crd_instance.to_dict()),
                self._extended_client.serialize(mutated_crd.to_dict()),
            ).patch
            self.logger.debug(f"crd_instance: {crd_instance.to_dict()}")
            self.logger.debug(f"mutated_crd: {mutated_crd.to_dict()}")
            self.logger.debug(f"patch operations: {patch_ops}")
            patch_b64 = base64.b64encode(json.dumps(patch_ops).encode("utf-8")).decode(
                "utf-8"
            )

            self.logger.debug("mutation passed")
            response = {
                "apiVersion": "admission.k8s.io/v1",
                "kind": "AdmissionReview",
                "response": {
                    "uid": r.get("uid"),
                    "allowed": True,
                    "patchType": "JSONPatch",
                    "patch": (patch_b64),
                },
            }
            return (
                json.dumps(response),
                falcon.HTTP_200,
                {"Content-Type": "application/json"},
            )
        except MutationWebhookError as e:
            self.logger.warning(f"mutation failed: {e.reason}")
            response = {
                "apiVersion": "admission.k8s.io/v1",
                "kind": "AdmissionReview",
                "response": {
                    "uid": r.get("uid") if request else None,
                    "allowed": False,
                    "status": {"message": e.reason},
                },
            }
            return (
                json.dumps(response),
                falcon.HTTP_200,
                {"Content-Type": "application/json"},
            )



import logging
import sys


FMT = 'timestamp=%(asctime)s name=%(name)s level=%(levelname)s msg="%(message)s"'


root_logger = logging.getLogger()
formater = logging.Formatter(FMT)

stdout_handler = logging.StreamHandler(stream=sys.stdout)
stdout_handler.setFormatter(formater)

root_logger.addHandler(stdout_handler)


from inspect import isclass
import re
from typing import Any, Dict, TypeVar, cast, get_args, get_origin
from kubernetes import client

T = TypeVar("T")

NATIVE_TYPES = {
    "str": "string",
    "int": "integer",
    "float": "number",
    "dict": "object",
    "bool": "boolean",
    "bytes": "byte",
    "datetime": "string",
    "object": "object",
}


def parse_kube_prop(name: str, klass: type | str):
    """
    Recursively parse a kubernetes.client model class (e.g., V1ObjectMeta) or a string type
    into a PropYaml, handling lists and nested models.
    """
    # Handle string type hints like 'list[V1Container]'
    if isinstance(klass, str):
        if klass.startswith("list["):
            sub_type_str = re.match(r"list\[(.*)\]", klass).group(1)  # type: ignore
            # Try to resolve the class from the client module
            sub_type = getattr(client, sub_type_str, sub_type_str)
            subtype_yaml = parse_kube_prop(name, sub_type)
            return PropYaml(
                typ="array",
                subtype=subtype_yaml.typ,
                subtype_props=subtype_yaml.subprops,
            )
        if klass.startswith("dict("):
            sub_type_str = re.match(r"dict\((.*)\)", klass).group(1)  # type: ignore
            # Try to resolve the class from the client module
            sub_type = getattr(client, sub_type_str, sub_type_str)
            return PropYaml(typ="object")
        # Try to resolve the class from the client module
        klass_obj = getattr(client, klass, None)
        if klass_obj and isclass(klass_obj):
            return parse_kube_prop(name, klass_obj)
        # Fallback: treat as primitive

        return PropYaml(typ=NATIVE_TYPES[klass])

    if klass in NATIVE_TYPES:
        return PropYaml(typ=NATIVE_TYPES[klass])

    # Handle kubernetes client models
    if hasattr(klass, "openapi_types") and hasattr(klass, "attribute_map"):
        properties = {}
        for attr, attr_type in getattr(klass, "openapi_types").items():
            prop_yaml = parse_kube_prop(attr, attr_type)
            # Use the attribute_map to get the serialized name
            yaml_name = klass.attribute_map[attr]
            properties[yaml_name] = prop_yaml
        desc = klass.__doc__.strip() if klass.__doc__ else None
        return PropYaml(
            typ="object",
            properties=properties,
            subtype_desc=desc,
        )

    # Fallback: treat as string
    print(name, klass)
    
    return PropYaml(typ="string")


def prop(
    typ: type[T],
    required=False,
    properties: dict[str, Any] | None = None,
    **kwargs: Any,
) -> T:
    """
    Define a propertie of a CRD, the available types are
    `str`, `int`, `float`, `dict`, `bool`, `list[Any]`,
    subclasses of `OpenAPISchema` and classes from `kubernetes.client`
    """
    t = NATIVE_TYPES.get(typ.__name__, None)
    subtype = None
    subprops = None
    subtype_desc = None
    if isclass(typ) and hasattr(typ, "openapi_types") and hasattr(typ, "attribute_map"):
        if properties is not None:
            raise RuntimeError("a class-prop cannot have properties defined in it")
        t = "object"
        properties = {}
        if typ != client.V1ObjectMeta:
            for attr, attr_type in cast(
                Dict[str, Any], getattr(typ, "openapi_types")
            ).items():
                properties[getattr(typ, "attribute_map")[attr]] = (
                    getattr(typ, attr)
                    if isclass(attr_type) or get_origin(attr_type) is list
                    else parse_kube_prop(attr, attr_type)
                )

            if typ.__doc__ is not None:
                kwargs["description"] = typ.__doc__.strip()
    if t is None:
        if get_origin(typ) is list:
            t = "array"
            sub_kls = get_args(typ)[0]
            subtype = NATIVE_TYPES.get(sub_kls.__name__, None)
            if isclass(sub_kls) and hasattr(sub_kls, "openapi_types"):
                subtype = "object"
                subtyp = sub_kls
                subprops = {}
                for attr, attr_type in cast(
                    Dict[str, str], getattr(subtyp, "openapi_types")
                ).items():
                    print(attr, attr_type, get_origin(attr_type) is list)
                    subprops[getattr(subtyp, "attribute_map")[attr]] = (
                        getattr(subtyp, attr)
                        if isclass(attr_type) or get_origin(attr_type) is list
                        else parse_kube_prop(attr, attr_type)
                    )

                if subtyp.__doc__ is not None:
                    subtype_desc = subtyp.__doc__.strip()

    if t is None or (t == "array" and subtype is None):
        supported_types = "`, `".join(list(NATIVE_TYPES))
        raise TypeError(
            f"`{typ}` not suported",
            f"`{supported_types}`",
            "kubernetes.client classes and",
            "and lists of these types are allowed",
        )

    p = PropYaml(
        typ=t,
        required=required,
        properties=properties,
        subtype=subtype,
        subtype_props=subprops,
        subtype_desc=subtype_desc,
        **kwargs,
    )
    p.real_type = typ
    return cast(T, p)


class PropYaml:
    """
    The class that is mapped to YAML
    """

    typ: str
    required: bool
    args: dict
    subprops: dict | None
    subtype: str | None
    subtype_props: dict | None
    subtype_desc: str | None
    real_type: Any

    def __init__(
        self,
        typ: str,
        subtype: str | None = None,
        subtype_props: dict | None = None,
        subtype_desc: str | None = None,
        required: bool = False,
        properties: dict | None = None,
        **kwargs,
    ):
        self.typ = typ
        self.required = required
        self.subprops = properties
        self.subtype = subtype
        self.subtype_props = subtype_props
        self.subtype_desc = subtype_desc
        self.args = kwargs



import importlib
import inspect
import os
from pathlib import Path
import select
import subprocess
from typing import List
import caseconverter
import click
import tomlkit
from kuroboros.config import KuroborosConfig

from kuroboros.controller import ControllerConfig, ControllerConfigVersions
from kuroboros.exceptions import MultipleDefinitionsException
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.reconciler import BaseReconciler
from kuroboros.schema import CRDSchema
from kuroboros.webhook import BaseMutationWebhook, BaseValidationWebhook


def yaml_format(value):
    """Converts Python types to YAML-compatible strings with proper quoting"""
    if isinstance(value, bool):
        return "true" if value else "false"  # Handle booleans
    if value is None:
        return "null"  # Handle None
    if isinstance(value, (int, float)):
        return str(value)  # Numbers remain unquoted
    if isinstance(value, str):
        # Quote strings that are numeric or contain special characters
        try:
            # Check if string is numeric
            float(value)
            return f'"{value}"'  # Quote numeric-looking strings
        except ValueError:
            # Quote strings with colons, spaces, etc.
            if any(c in value for c in ":[]{}, "):
                return f'"{value}"'
            return value  # Unquoted for simple strings
    else:
        return str(value)  # Fallback for other types


def x_kubernetes_kebab(name: str) -> str:
    """
    Handle special cases for x_kubernetes_ prop prop names
    """
    if name.startswith("x_kubernetes_"):
        return caseconverter.kebabcase(name)

    return name


def create_file(
    output: str, file_name: str, data: str, overwrite: bool = True, parents: bool = True
):
    """
    Creates a file in a given path
    """
    p = Path(f"{output}/{file_name}")
    p.parent.mkdir(parents=parents, exist_ok=True)
    if p.is_file() and not overwrite:
        click.echo(f"not overwriten: {file_name}.")
        return
    try:
        action = ""
        if p.is_file():
            action = "overwriten"
        else:
            action = "created"
        with open(f"{output}/{file_name}", "w", encoding="utf-8") as file:
            file.write(data)
            file.close()

        click.echo(f"{action}: {file_name}")

    except Exception as e:
        click.echo(f"error while craeting file {output}")
        raise e


def run_command(command) -> str:
    """
    Runs a command and return the ouput once finished
    """
    result = subprocess.run(
        command,
        check=True,
        text=True,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    return result.stdout


def get_image_info():
    """
    Gets the operator image from the KuroborosConfig
    """
    reg = KuroborosConfig.get("build", "image", "registry", typ=str)
    repo = KuroborosConfig.get("build", "image", "repository", typ=str)
    tag_suffix = KuroborosConfig.get("build", "image", "tag_suffix", typ=str)
    tag_prefix = KuroborosConfig.get("build", "image", "tag_prefix", typ=str)
    tag = KuroborosConfig.get("build", "image", "tag", typ=str)
    if tag in ("$PYPROJECT_VERSION", "$TAG_OR_SHORT_COMMIT_SHA"):
        if tag == "$PYPROJECT_VERSION":
            with open("./pyproject.toml", "r", encoding="utf-8") as pyproject:
                content = pyproject.read()
                pyproject_toml = tomlkit.loads(content)
                tag = pyproject_toml.get("project", {}).get("version", "latest")
        else:
            tag = run_command(
                "git describe --exact-match --tags 2> /dev/null || git rev-parse --short HEAD"
            )

    img = f"{repo}:{tag_prefix}{tag}{tag_suffix}"
    if reg != "":
        img = f"{reg}/{img}"

    return img


def run_command_stream_simple(command, env=None):
    """
    Runs a siumple shell command and stream the output
    """
    if env is None:
        env = {}
    print(f"running command: {command}")
    with subprocess.Popen(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        env={**os.environ, **env},
        bufsize=1,
        universal_newlines=True,
    ) as process:

        stdout = []
        stderr = []

        assert process.stdout is not None
        assert process.stderr is not None
        fd_stdout = process.stdout.fileno()
        fd_stderr = process.stderr.fileno()

        while True:
            # Check stdout
            reads, _, _ = select.select([fd_stdout, fd_stderr], [], [], 0.1)
            if fd_stdout in reads:
                line = process.stdout.readline()
                if line:
                    print(line, end="", flush=True)  # Force flush
                    stdout.append(line)

            if fd_stderr in reads:
                line = process.stderr.readline()
                if line:
                    print(line, end="", flush=True)  # Force flush
                    stderr.append(line)

            # Check if process has finished
            if process.poll() is not None:
                # Read any remaining output
                for line in process.stdout:
                    print(line, end="", flush=True)
                    stdout.append(line)
                for line in process.stderr:
                    print(line, end="", flush=True)
                    stderr.append(line)
                break


def load_controller_configs(controllers_path) -> List[ControllerConfig]:
    """
    Loads the controller config given the controllers path, by default is /controllers
    """
    controllers_configs: List[ControllerConfig] = []
    path = os.path.join(Path().absolute(), controllers_path)
    directory = Path(path)
    try:
        # each folder in /controllers
        controllers = [entry.name for entry in directory.iterdir() if entry.is_dir()]
    except Exception:  # pylint: disable=broad-except
        controllers = []
    for controller in controllers:
        # we assume that each controller has a group_version.py file
        # and a versions folder with the versions of the controller
        ctrl_conf = ControllerConfig()
        ctrl_conf.name = controller
        try:
            group_version_module = importlib.import_module(
                f"{controllers_path}.{controller}.group_version"
            )
        except ModuleNotFoundError:
            # If we dont find any GVI we skip this controller, as its not a controller (?)
            continue
        group_version = None
        for _, obj in inspect.getmembers(group_version_module):
            if isinstance(obj, GroupVersionInfo):
                group_version = obj

        if group_version is None:
            # If we find a group_versin.py file but it doesn't contain any GVI object
            continue
        ctrl_conf.group_version_info = group_version
        versions_path = os.path.join(path, controller)
        versions_dir = Path(versions_path)
        versions = [
            entry.name
            for entry in versions_dir.iterdir()
            if entry.is_dir() and GroupVersionInfo.is_valid_api_version(entry.name)
        ]
        for version in versions:
            # each version folder should have python files with the reconciler, crd classes and
            # posibly validation and mutation webhook
            ctrl_versions = ControllerConfigVersions()
            ctrl_versions.name = version
            modules = [
                {"name": "crd", "required": True},
                {"name": "reconciler", "required": True},
                {"name": "mutation", "required": False},
                {"name": "validation", "required": False},
            ]

            for module_conf in modules:
                module_name = module_conf["name"]
                module = None
                try:
                    module = importlib.import_module(
                        f"{controllers_path}.{controller}.{version}.{module_name}"
                    )
                except ModuleNotFoundError as e:
                    if module_conf["required"]:
                        raise e
                    continue

                assert module is not None
                for _, obj in inspect.getmembers(module, inspect.isclass):
                    if (
                        module_name == "reconciler"
                    ):  # Load reconciler from reconciler.py
                        if BaseReconciler in obj.__bases__:
                            if ctrl_versions.reconciler is not None:
                                raise MultipleDefinitionsException(
                                    ctrl_versions.reconciler, controller, version
                                )
                            obj.set_gvi(group_version)
                            ctrl_versions.reconciler = obj
                    elif module_name == "crd":  # Load CRD from crd.py
                        if CRDSchema in obj.__bases__:
                            if ctrl_versions.crd is not None:
                                raise MultipleDefinitionsException(
                                    ctrl_versions.crd, controller, version
                                )

                            obj.set_gvi(group_version)
                            ctrl_versions.crd = obj
                    elif (
                        module_name == "validation"
                    ):  # Load validation webhook from validation.py
                        if BaseValidationWebhook in obj.__bases__:
                            if ctrl_versions.validation_webhook is not None:
                                raise MultipleDefinitionsException(
                                    ctrl_versions.validation_webhook,
                                    controller,
                                    version,
                                )
                            obj.set_gvi(group_version)
                            ctrl_versions.validation_webhook = obj
                    elif (
                        module_name == "mutation"
                    ):  # Load mutation webhook from validation.py
                        if BaseMutationWebhook in obj.__bases__:
                            if ctrl_versions.mutation_webhook is not None:
                                raise MultipleDefinitionsException(
                                    ctrl_versions.mutation_webhook, controller, version
                                )
                            obj.set_gvi(group_version)
                            ctrl_versions.mutation_webhook = obj
                # Only append after all fields are set and only if valid
                if (
                    ctrl_versions.reconciler is not None
                    and ctrl_versions.crd is not None
                ):
                    ctrl_conf.versions.append(ctrl_versions)

        # If the controller has at least 1 valid version we append it, otherwhise we ignore it
        if len(ctrl_conf.versions) > 0:
            controllers_configs.append(ctrl_conf)

    return controllers_configs



from typing import Any, Dict, List

import falcon
from gunicorn import glogging
from gunicorn.app.base import BaseApplication

from kuroboros import logger
from kuroboros.webhook import BaseWebhook



class InjectedLogger(glogging.Logger):
    """
    The injected logger to the GunicornApp
    """
    def setup(self, cfg):
        logg = logger.root_logger.getChild(__name__)
        self.error_log = logg.getChild("gunicorn")
        self.access_log = logg.getChild("gunicorn.access")


class GunicornApp(BaseApplication):
    """
    The GunicornApp that runs the Falcon app
    """
    def __init__(self, app, options=None):
        self.application = app
        self.options = options or {}
        super().__init__()

    def init(self, parser, opts, args):
        # No additional initialization required
        return None

    def load_config(self):
        if self.cfg is not None:
            for key, value in self.options.items():
                self.cfg.set(key.lower(), value)

            self.cfg.set("logger_class", InjectedLogger)

    def access(self, res, req, environ, req_time):
        """
        Access logs
        """
        self.application.access_log.info(
            f"{req.method} {req.path} {res.status} {req_time:.3f}s "
            f"{environ.get('REMOTE_ADDR', 'unknown')}"
        )

    def load(self):
        return self.application


class HTTPSWebhookServer:
    """
    A Falcon app that runs in a Gunicorn app with SSL. 
    This HTTP server will register all webhooks defined in the project controllers
    """
    port: int
    host: str
    _endpoints: List[BaseWebhook]
    _falcon: falcon.App
    _server: GunicornApp
    _server_options: Dict[str, Any]
    _logger = logger.root_logger.getChild(__name__)

    def __init__(
        self,
        cert_file: str,
        key_file: str,
        endpoints: List[BaseWebhook],
        port: int = 443,
        host: str = "0.0.0.0",
        workers: int = 4,
    ) -> None:
        self.port = port
        self.host = host

        self._falcon = falcon.App()
        self._endpoints = endpoints
        self._server_options = {
            "bind": f"{self.host}:{self.port}",
            "workers": workers,
            "certfile": cert_file,
            "keyfile": key_file,
            "worker_class": "sync",
        }
        self._server = GunicornApp(self._falcon, self._server_options)

    def start(self) -> None:
        """
        Adds all the routes to the Falcon app and starts it
        """
        self._logger.info(f"starting webhook server on {self.host}:{self.port}")
        self._logger.info(f"using cert file: {self._server_options.get('certfile')}")
        self._logger.info(f"using key file: {self._server_options.get('keyfile')}")
        for webhook in self._endpoints:
            self._logger.info(
                f"registering endpoint: {webhook.name} at {webhook.endpoint}"
            )
            self._falcon.add_route(webhook.endpoint, webhook)

        self._server.run()



from logging import Logger
from typing import Any, Dict, Generator, List, MutableMapping, Tuple, Type, cast
import time
import threading
import caseconverter
from kubernetes import client, watch

from kuroboros import logger
from kuroboros.config import KuroborosConfig
from kuroboros.extended_api import ExtendedApi, ExtendedClient
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.reconciler import BaseReconciler
from kuroboros.schema import CRDSchema
from kuroboros.utils import NamespaceName, event_aware_sleep
from kuroboros.webhook import BaseMutationWebhook, BaseValidationWebhook


class EventEnum:
    """
    Event types from Kubernetes for the Controller
    """

    ADDED = "ADDED"
    MODIFIED = "MODIFIED"
    DELETED = "DELETED"


class ControllerConfigVersions:
    """
    The version configuration for the ControllerConfig
    """

    name: str
    reconciler: Type[BaseReconciler] | None = None
    crd: Type[CRDSchema] | None = None
    validation_webhook: Type[BaseValidationWebhook] | None = None
    mutation_webhook: Type[BaseMutationWebhook] | None = None

    def has_webhooks(self) -> bool:
        """
        Checks if any webhook is loaded
        """
        return self.validation_webhook is not None or self.mutation_webhook is not None


class ControllerConfig:
    """
    Configuration for the Controller
    """

    name: str
    group_version_info: GroupVersionInfo
    versions: List[ControllerConfigVersions]

    def __init__(self) -> None:
        self.versions = []

    def has_webhooks(self) -> bool:
        """
        Checks if the running version has webhooks loaded
        """
        return self.get_run_version().has_webhooks()

    @property
    def validation_webhook(self) -> Type[BaseValidationWebhook] | None:
        """
        Returns the validation webhook for the current version
        """
        return self.get_run_version().validation_webhook

    @property
    def mutation_webhook(self) -> Type[BaseMutationWebhook] | None:
        """
        Returns the mutation webhook for the current version
        """
        return self.get_run_version().mutation_webhook

    def get_run_version(self):
        """
        Gets the version that match the GroupVersionInfo
        """
        for version in self.versions:
            if self.group_version_info.api_version == version.name:
                return version
        raise RuntimeError(f"no version match {self.group_version_info.api_version}")


class Controller:
    """
    Kuroboros Controller class.
    Triggers the reconciliation_loop of CR and manage the Threads
    """

    _cleanup_interval: float
    _logger: Logger
    _members: MutableMapping[NamespaceName, BaseReconciler]
    _pending_remove: List[NamespaceName]
    _group_version_info: GroupVersionInfo
    _stop: threading.Event
    _watcher: watch.Watch
    _watcher_loop: threading.Thread
    _cleanup_loop: threading.Thread
    _api_client: ExtendedClient
    _extended_api: ExtendedApi

    reconciler: Type[BaseReconciler]
    validation_webhook: BaseValidationWebhook | None
    mutation_webhook: BaseMutationWebhook | None
    name: str

    @property
    def threads(self) -> int:
        """
        Returns the number of currently watched `Threads` of the controller
        """
        return len(self._members)

    def __init__(
        self,
        name: str,
        group_version_info: GroupVersionInfo,
        reconciler: Type[BaseReconciler],
        validation_webhook: Type[BaseValidationWebhook] | None = None,
        mutation_webhook: Type[BaseMutationWebhook] | None = None,
    ) -> None:

        if (
            validation_webhook is not None
            and reconciler.crd_type() != validation_webhook.crd_type()
        ):
            raise RuntimeError(
                "The validation webhook type must match the reconciler type",
                f"{reconciler.crd_type()} != {validation_webhook.crd_type()}",
            )

        self._cleanup_interval = KuroborosConfig.get(
            "operator", "cleanup_interval_seconds", typ=float
        )
        self._group_version_info = group_version_info
        pascal_name = caseconverter.pascalcase(name)
        self.name = f"{pascal_name}{group_version_info.pversion()}Controller"
        self._logger = logger.root_logger.getChild(__name__).getChild(self.name)
        self._check_permissions()
        self.reconciler = reconciler
        self._api_client = ExtendedClient()
        self._members = {}
        self._pending_remove = []
        self.validation_webhook = (
            validation_webhook() if validation_webhook is not None else None
        )
        self.mutation_webhook = (
            mutation_webhook() if mutation_webhook is not None else None
        )
        self._stop = threading.Event()

    def _check_permissions(self):
        api = client.AuthorizationV1Api()
        for verb in ["create", "list", "watch", "delete", "get", "patch", "update"]:
            resource_attributes = client.V1ResourceAttributes(
                group=self._group_version_info.group,
                resource=self._group_version_info.plural,
                verb=verb,
            )

            access_review = client.V1SelfSubjectAccessReview(
                spec=client.V1SelfSubjectAccessReviewSpec(
                    resource_attributes=resource_attributes
                )
            )

            res = api.create_self_subject_access_review(access_review)
            response = cast(client.V1SelfSubjectAccessReview, res)

            if response.status is not None and response.status.allowed:
                continue

            if response.status is not None and response.status.denied:
                crd_name = self._group_version_info.crd_name
                raise RuntimeWarning(
                    f"operator doesn't have {verb} permission over the CRD {crd_name}"
                )

    def _add_member(self, namespace_name: NamespaceName):
        """
        Adds the object to be managed and starts its `_reconcile` function
        in a thread
        """
        if namespace_name not in self._members:
            reconciler = self.reconciler(namespace_name)
            reconciler.start(extended_api=self._extended_api)
            self._members[namespace_name] = reconciler
            self._logger.info(
                "%s added as member",
                reconciler,
            )
        else:
            if self._members[namespace_name].is_running():
                return
            self._members[namespace_name].start(extended_api=self._extended_api)

    def _add_pending_remove(self, namespace_name: NamespaceName):
        """
        Adds the object to be safely removed from the management list
        """
        if namespace_name in self._pending_remove:
            return
        self._pending_remove.append(namespace_name)
        self._logger.info(
            "%s CR added as pending_remove",
            self._group_version_info.pkind(namespace_name),
        )

    def _remove_member(self, namespace_name: NamespaceName):
        """
        Sends an stop event to the member thread and stops the loop
        """
        if namespace_name not in self._members:
            return
        self._members[namespace_name].stop()
        self._members.pop(namespace_name)
        self._logger.info(
            "%s CR removed",
            self._group_version_info.pkind(namespace_name),
        )

    def _get_current_cr_list(self) -> List[Any]:
        """
        Gets the current list of objects in the cluster
        """
        return self._extended_api.get(
            kind=self._group_version_info.kind,
            api_version=f"{self._group_version_info.group}/{self._group_version_info.api_version}",
            klass=list[object],
        )

    def _stream_events(
        self,
        watcher: watch.Watch,
    ) -> Generator[Any | dict | str, str, None]:
        """
        Wrapper to `kubernetes.watch.Watch().stream()`
        """
        cr_api = client.CustomObjectsApi()
        return watcher.stream(
            cr_api.list_cluster_custom_object,
            group=self._group_version_info.group,
            version=self._group_version_info.api_version,
            plural=self._group_version_info.plural,
        )

    def _preload_existing_cr(self):
        self._logger.info(
            "preloading existing %s CRs", self._group_version_info.pkind()
        )
        try:
            current_crs = self._get_current_cr_list()
            for pending in current_crs:
                crd_namespace_name = (
                    pending["metadata"].get("namespace", None),
                    pending["metadata"]["name"],
                )
                self._add_member(crd_namespace_name)
            self._logger.info(
                "preloaded %d %s CR(s)",
                len(current_crs),
                self._group_version_info.pkind(),
            )
        except Exception as e:
            self._logger.error(
                "error while preloading %s CR: %s",
                self._group_version_info.pkind(),
                e,
                exc_info=True,
            )
            raise e

    def _watch_pending_remove(self):
        """
        Looks for the objects with `finalizers` pending to be removed
        every 5 seconds and removes them once they no longer exists
        """
        self._logger.info(
            "watching %s CRs pending to remove",
            self._group_version_info.pkind(),
        )
        while not self._stop.is_set():
            for namespace, name in self._pending_remove:
                self._logger.info(
                    "%d %s CRs pending to remove",
                    len(self._pending_remove),
                    self._group_version_info.pkind(),
                )

                try:
                    if not self._cr_exists(name, namespace):
                        self._remove_member((namespace, name))
                        self._pending_remove.remove((namespace, name))
                        self._logger.info(
                            "%s CR no longer found, removed",
                            self._group_version_info.pkind((namespace, name)),
                        )
                except client.ApiException as e:
                    self._logger.error(
                        "unexpected api error ocurred while watching pending_remove %s CR: %s",
                        self._group_version_info.pkind((namespace, name)),
                        e,
                        exc_info=True,
                    )
            event_aware_sleep(self._stop, self._cleanup_interval)

    def _watch_cr_events(self):
        """
        Watch for the kubernetes events of the object.
        Adds the member if its `ADDED` or `MODIFIED` and removes them when `DELETED`
        """
        self._logger.info(
            "starting to watch %s events", self._group_version_info.pkind()
        )
        self._watcher = watch.Watch()
        try:
            for event in self._stream_events(self._watcher):
                if self._stop.is_set():
                    self._watcher.stop()
                    break
                if not isinstance(event, dict):
                    self._logger.warning("event received is not a dict, skipping")
                    continue
                try:
                    e_type = event["type"]
                    raw_cr = event["object"]
                    metadata: Dict[str, Any] = raw_cr["metadata"]
                    namespace_name = (
                        metadata.get("namespace", None),
                        metadata["name"],
                    )
                    finalizers = metadata.get("finalizers")
                    self._logger.debug(f"event: {namespace_name} {event['type']}")
                    if e_type in (EventEnum.ADDED, EventEnum.MODIFIED):
                        self._add_member(namespace_name)
                    elif e_type == EventEnum.DELETED:
                        if finalizers is not None and len(finalizers) > 0:
                            self._add_pending_remove(namespace_name)
                            continue
                        self._remove_member(namespace_name)
                    else:
                        self._logger.warning(f"event type {event['type']} not handled")

                except Exception as e:  # pylint: disable=broad-exception-caught
                    self._logger.warning(
                        "an Exception ocurred while streaming %s events: %s",
                        self._group_version_info.pkind(),
                        e,
                        exc_info=True,
                    )
                    continue
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._logger.error(
                "error while watching %s: %s",
                self._group_version_info.pkind(),
                e,
                exc_info=True,
            )
        finally:
            self._logger.info(
                "no longer watching events from %s",
                self._group_version_info.pkind(),
            )
            self._watcher.stop()

    def _cr_exists(self, name: str, namespace: str | None = None) -> bool:
        try:
            group = self._group_version_info.group
            api_version = self._group_version_info.api_version
            self._extended_api.get(
                kind=self._group_version_info.kind,
                api_version=f"{group}/{api_version}",
                name=name,
                namespace=namespace,
            )
        except client.ApiException as e:
            if e.status == 404:
                return False
            raise e
        return True

    def run(self) -> Tuple[threading.Thread, threading.Thread]:
        """
        Starts the controller.
        Pre-load the current CRs in the cluster ands start both watcher and cleanup threads.
        """
        self._extended_api = ExtendedApi(api_client=self._api_client)
        self._watcher_loop = threading.Thread(
            target=self._watch_cr_events, name=f"{self.name}::Watcher", daemon=True
        )
        self._cleanup_loop = threading.Thread(
            target=self._watch_pending_remove, name=f"{self.name}::Cleanup", daemon=True
        )

        self._preload_existing_cr()

        self._watcher_loop.start()
        self._cleanup_loop.start()

        return (self._watcher_loop, self._cleanup_loop)

    def stop(self):
        """
        Send a stop event to every thread under this controller and waits for every CR to react
        to the event
        """
        self._logger.info("stopping controller %s", self.name)
        if self._watcher is not None:
            self._watcher.stop()
        if not self._stop.is_set():
            self._stop.set()

        wait_for_stop: List[BaseReconciler] = []
        for namespace_name, member in self._members.items():
            self._logger.debug(
                "sending stop event to %s",
                self._group_version_info.pkind(namespace_name),
            )
            member.stop()
            wait_for_stop.append(member)

        alive_threads = wait_for_stop
        if len(alive_threads) > 0:
            self._logger.debug(
                "waiting for %d reconcilation loop(s) thread(s) to stop...",
                len(alive_threads),
            )
            while len(alive_threads) > 0:
                alive_threads = [
                    member for member in wait_for_stop if member.is_running()
                ]
                time.sleep(0.5)

        self._logger.info("controller %s succesfully stopped", self.name)



import pprint
from typing import Any, ClassVar, Dict, get_origin, get_args

import caseconverter

from kuroboros.schema.properties import PropYaml


class OpenAPISchema:
    """
    A Base OpenAPI Model to define CRD Properties.
    It auotogenerates attribuite_map and openapi_types based on props
    """

    attribute_map: ClassVar[Dict[str, str]]
    openapi_types: ClassVar[Dict[str, Any]]
    _data: dict

    def __init__(self, **kwargs) -> None:
        self.load_data(**kwargs)

    def __init_subclass__(cls) -> None:
        ## populate the attribute_map and openapi_types
        if "openapi_types" not in cls.__dict__:
            cls.openapi_types = {}

        if "attribute_map" not in cls.__dict__:
            cls.attribute_map = {}

        for attribute, value in cls.__dict__.items():
            if (
                attribute[:2] != "__"
                and not callable(value)
                and isinstance(value, PropYaml)
            ):
                cls.openapi_types[attribute] = value.real_type
                cls.attribute_map[attribute] = caseconverter.camelcase(attribute)

    def __getattribute__(self, name: str):
        attr = object.__getattribute__(self, name)
        data = None
        try:
            data = object.__getattribute__(self, "_data")
        except AttributeError:
            data = {}
        try:
            if isinstance(attr, PropYaml):
                return data[name]
            return attr
        except KeyError:
            return None

    def __setattr__(self, name, value):
        # If setting a property, update both self._data and parent if present
        attr = object.__getattribute__(self, name)
        if isinstance(attr, PropYaml):
            self._data[name] = value
        else:
            object.__setattr__(self, name, value)

    def __repr__(self) -> str:
        return pprint.pformat(self.to_dict())

    def load_data(self, **kwargs):
        """
        load data given given the Schema attributes in attribute_map keys
        """
        data = {}
        for arg, val in kwargs.items():
            if arg not in self.attribute_map:
                continue
            propertie = object.__getattribute__(self, arg)
            if isinstance(propertie, PropYaml):
                prop_type = propertie.real_type
                if get_origin(prop_type) is list:
                    data[arg] = [
                        (
                            el
                            if hasattr(el, "openapi_types")
                            else get_args(prop_type)[0](**el)
                        )
                        for el in val
                    ]

                elif hasattr(prop_type, "openapi_types") and val is not None:
                    data[arg] = (
                        prop_type(**val.to_dict())
                        if hasattr(val, "to_dict")
                        else prop_type(**val)
                    )
                else:
                    data[arg] = val
        object.__setattr__(self, "_data", data)

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr in self.attribute_map:
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (
                            (item[0], item[1].to_dict())
                            if hasattr(item[1], "to_dict")
                            else item
                        ),
                        value.items(),
                    )
                )
            else:
                result[attr] = value
        return result



import logging
import multiprocessing
import signal
import sys
import threading
import time
from typing import Dict, List, Tuple, Type, cast
import uuid

import caseconverter
from kubernetes import client, config
from prometheus_client import Gauge, start_http_server

from kuroboros import logger
from kuroboros.config import OPERATOR_NAMESPACE, KuroborosConfig
from kuroboros.controller import Controller, ControllerConfig
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.reconciler import BaseReconciler
from kuroboros.utils import event_aware_sleep
from kuroboros.webhook import BaseMutationWebhook, BaseValidationWebhook, BaseWebhook
from kuroboros.webhook_server import HTTPSWebhookServer


class Operator:
    """
    The kuroboros Operator, it collect metrics, acquiere leadership,
    start the webhook server and start the controllers reconcilation loops
    """

    _leader_interval: float
    _metrics_interval: float
    _metrics_port: int
    _cert_path: str
    _key_path: str
    _webhook_port: int
    _gunicorn_workers: int
    _running: bool
    _uid: str
    _logger: logging.Logger
    _is_leader: threading.Event
    _threads_by_reconciler: Gauge
    _python_threads: Gauge
    _stop: threading.Event
    _threads: List[threading.Thread]
    _webhook_server_process: multiprocessing.Process | None
    _interrupted = False

    _namespace: str
    _controllers: List[Controller]
    _controller_threads: Dict[Controller, Tuple[threading.Thread, threading.Thread]] = (
        {}
    )

    name: str

    def __init__(self) -> None:
        self.name = KuroborosConfig.get("operator", "name", typ=str)
        self._leader_interval = KuroborosConfig.get(
            "operator", "leader_acquire_interval_seconds", typ=float
        )
        self._metrics_interval = KuroborosConfig.get(
            "operator", "metrics", "interval_seconds", typ=float
        )
        self._metrics_port = KuroborosConfig.get("operator", "metrics", "port", typ=int)
        self._cert_path = KuroborosConfig.get(
            "operator", "webhook_server", "cert_path", typ=str
        )
        self._key_path = KuroborosConfig.get(
            "operator", "webhook_server", "key_path", typ=str
        )
        self._gunicorn_workers = KuroborosConfig.get(
            "operator", "webhook_server", "gunicorn_workers", typ=int
        )
        self._webhook_port = KuroborosConfig.get(
            "operator", "webhook_server", "port", typ=int
        )
        self._threads_by_reconciler = Gauge(
            "kuroboros_python_threads_by_reconciler",
            "The number of threads running by the CRD controller",
            labelnames=["namespace", "reconciler"],
        )
        self._python_threads = Gauge(
            "python_active_threads", "The number of active python threads"
        )
        self._controllers = []
        self._threads = []
        self._webhook_server_process = None
        self._running = False
        self._is_leader = threading.Event()
        self._uid = str(uuid.uuid4())
        self._namespace = OPERATOR_NAMESPACE
        self._logger = logger.root_logger.getChild(__name__).getChild(
            caseconverter.pascalcase(self.name)
        )
        self._stop = threading.Event()
        try:
            config.load_kube_config()
        except Exception:  # pylint: disable=broad-except
            config.load_incluster_config()

    def is_leader(self) -> bool:
        """
        Check if the leader event is set
        """
        return self._is_leader.is_set()

    def is_running(self) -> bool:
        """
        Checks that the operator is running
        """
        return self._running

    @property
    def namespace(self) -> str:
        """
        Get the opeartor namespace (if any)
        """
        return self._namespace

    @property
    def uid(self) -> str:
        """
        Get the random UID generated for the Opeartor
        """
        return self._uid

    @property
    def controllers(self) -> List[Controller]:
        """
        Get a copy of the Controllers in tyhe Operator
        """
        return self._controllers.copy()

    def _add_controller(
        self,
        name: str,
        group_version: GroupVersionInfo,
        reconciler: Type[BaseReconciler],
        validation_webhook: Type[BaseValidationWebhook] | None = None,
        mutation_webhook: Type[BaseMutationWebhook] | None = None,
    ) -> None:
        if self.is_running():
            raise RuntimeError("cannot add controller while operator is running")

        controller = Controller(
            name=name,
            group_version_info=group_version,
            reconciler=reconciler,
            validation_webhook=validation_webhook,
            mutation_webhook=mutation_webhook,
        )
        if controller.name in [ctrl.name for ctrl in self._controllers]:
            raise RuntimeError("cannot add an already added controller")

        self._controllers.append(controller)

    def _acquire_leader_lease(self):
        api = client.CoordinationV1Api()
        lease_name = f"{caseconverter.kebabcase(self.name)}-leader"
        lease_duration = 10
        self._logger.info(f"trying to acquire leadership with uid: {self._uid}")
        while not self._stop.is_set():
            try:
                lease = api.read_namespaced_lease(
                    name=lease_name, namespace=self._namespace
                )
            except client.ApiException as e:
                if e.status == 404:
                    lease = client.V1Lease(
                        metadata=client.V1ObjectMeta(name=lease_name),
                        spec=client.V1LeaseSpec(
                            renew_time=time.strftime(
                                "%Y-%m-%dT%H:%M:%S.000000Z", time.gmtime()
                            ),
                            lease_duration_seconds=lease_duration,
                            holder_identity=self._uid,
                        ),
                    )
                    api.create_namespaced_lease(namespace=self._namespace, body=lease)
                    if not self.is_leader():
                        self._logger.info(f"leadership acquired under uid {self._uid}")
                        self._is_leader.set()

                else:
                    self._logger.error(
                        f"error while trying to acquire leadership lease: {e}",
                        exc_info=True,
                    )
                    raise RuntimeError("Error while acquiring leadership") from e
            lease_data: client.V1Lease = cast(client.V1Lease, lease)
            if lease_data.spec is None:
                raise RuntimeError("Unexpected empty lease.spec")
            current_time = time.time()
            renew_time = lease_data.spec.renew_time.timestamp()
            lease_expired = (
                current_time > renew_time + lease_data.spec.lease_duration_seconds
            )
            if lease_expired or lease_data.spec.holder_identity == self._uid:
                lease_data.spec.renew_time = time.strftime(
                    "%Y-%m-%dT%H:%M:%S.000000Z", time.gmtime()
                )
                lease_data.spec.holder_identity = self._uid
                lease_data.spec.lease_duration_seconds = lease_duration
                api.replace_namespaced_lease(
                    name=lease_name, namespace=self._namespace, body=lease_data
                )
                if not self._is_leader.is_set():
                    self._logger.info(f"leadership acquired under uid {self._uid}")
                    self._is_leader.set()

            time.sleep(self._leader_interval)

    def _metrics(self) -> None:

        while not self._stop.is_set():
            for ctrl in self._controllers:
                self._threads_by_reconciler.labels(
                    OPERATOR_NAMESPACE, ctrl.reconciler.__name__
                ).set(ctrl.threads)
            self._python_threads.set(threading.active_count())
            event_aware_sleep(self._stop, self._metrics_interval)

    def start(
        self,
        controllers: List[ControllerConfig],
        skip_controllers: bool = False,
        skip_webhook_server: bool = False,
    ) -> None:
        """
        Starts the opearator an the services after acquiring leadership of the CRs
        """

        if skip_controllers and skip_webhook_server:
            raise RuntimeError(
                "cannot start operator without running controllers or webhook server"
            )
        if self._running:
            raise RuntimeError("cannot start an already started Operator")

        if len(controllers) == 0:
            raise RuntimeError("no controllers found to run the operator")

        # Add Controllers from controller configs
        for ctrl in controllers:
            self._logger.debug(f"adding {ctrl.name} controller")
            run_version = ctrl.get_run_version()
            if run_version.reconciler is None:
                raise RuntimeError(
                    f"reconciler `None` in {ctrl.name} {run_version.name}"
                )

            try:
                self._add_controller(
                    name=ctrl.name,
                    group_version=ctrl.group_version_info,
                    reconciler=run_version.reconciler,
                    validation_webhook=run_version.validation_webhook,
                    mutation_webhook=run_version.mutation_webhook,
                )

            except RuntimeError as e:
                self._logger.warning(e)
                continue

        # Start the webhook server if needed
        if not skip_webhook_server:
            webhooks: List[BaseWebhook] = []
            for ctrl in self._controllers:
                if ctrl.validation_webhook is not None:
                    webhooks.append(ctrl.validation_webhook)
                if ctrl.mutation_webhook is not None:
                    webhooks.append(ctrl.mutation_webhook)

            if len(webhooks) > 0:
                webhook_server = HTTPSWebhookServer(
                    cert_file=self._cert_path,
                    key_file=self._key_path,
                    endpoints=webhooks,
                    port=self._webhook_port,
                    workers=self._gunicorn_workers
                )
                self._webhook_server_process = multiprocessing.Process(
                    target=webhook_server.start,
                    name=f"{self.name}-webhook-server-process",
                )
                self._webhook_server_process.start()

        # Start controllers
        if not skip_controllers:
            self._is_leader.clear()
            leader_election = threading.Thread(
                target=self._acquire_leader_lease,
                name=f"{self.name}::LeaderElection",
                daemon=True,
            )
            self._threads.append(leader_election)
            leader_election.start()
            while not self.is_leader():
                if not leader_election.is_alive():
                    raise RuntimeError(
                        "leader election loop died while trying to acquire leadership"
                    )
                continue

            for ctrl in self._controllers:
                ctrl_threads = ctrl.run()
                self._controller_threads[ctrl] = ctrl_threads

        # start metrics maybe
        try:
            start_http_server(self._metrics_port)
        except Exception:  # pylint: disable=broad-except
            self._logger.warning("metrics http server could not be started")

        metrics_loop = threading.Thread(
            target=self._metrics, name=f"{self.name}::Metrics", daemon=True
        )
        metrics_loop.start()
        self._threads.append(metrics_loop)
        self._running = True

        # handle stop
        signal.signal(signal.SIGINT, self.signal_stop)
        signal.signal(signal.SIGTERM, self.signal_stop)

        # handle crash
        while self._running:
            for thread in self._threads:
                if not thread.is_alive():
                    self._logger.error(f"Thread {thread.name} died unexpectedly")
                    raise threading.ThreadError("Death thread")

            if (
                self._webhook_server_process is not None
                and not self._webhook_server_process.is_alive()
            ):
                self._logger.error("Webhook Server Process died unexpectedly")
                raise multiprocessing.ProcessError("Death process")

            if not skip_controllers:
                for ctrl in self._controllers:
                    for thread in self._controller_threads[ctrl]:
                        if not thread.is_alive():
                            self._logger.error(
                                f"Controller {ctrl.name} thread {thread.name} died unexpectedly"
                            )
                            raise threading.ThreadError("Death thread")
            event_aware_sleep(self._stop, 1)

        self._logger.info("good bye!")

    def signal_stop(self, sig, _):
        """
        Starts the shutdown process on a SIGINT, sending stop events to every component and
        ppropagating the event to every thread
        """
        self._logger.warning(f"{signal.Signals(sig).name} received")
        if sig == 2:  # SIGINT
            if self._interrupted:
                self._logger.warning("second SIGINT received, killing process")
                sys.exit(1)
            self._interrupted = True
            self._logger.warning("trying to gracefully shutdown...")
            for ctrl in self._controllers:
                ctrl.stop()

            alive_threads = self._threads
            self._logger.debug(
                f"waiting for {len(alive_threads)} threads in the operator to stop..."
            )
            self._stop.set()
            while len(alive_threads) > 0:
                alive_threads = [
                    thread
                    for thread in self._threads
                    if thread.is_alive() and isinstance(thread, threading.Thread)
                ]
                time.sleep(0.5)

        self._running = False



from inspect import isclass
import json
from typing import Any, Type, TypeVar, cast, get_args, get_origin
from kubernetes import client, dynamic
from kuroboros.schema import CRDSchema
from kuroboros.schema import OpenAPISchema

R = TypeVar("R", bound=Any)


class ExtendedClient(client.ApiClient):
    """
    Implements `kubernetes.client.ApiClient` to add
    `OpenAPISchema` desearialization
    """

    def __deserialize_openapi(self, data, klass):
        kwargs = {}
        if (
            data is not None
            and klass.openapi_types is not None
            and isinstance(data, (list, dict))
        ):
            for attr, attr_type in klass.openapi_types.items():
                if klass.attribute_map[attr] in data:
                    value = data[klass.attribute_map[attr]]
                    kwargs[attr] = self.deserialize(value, attr_type)

        return klass(**kwargs)

    def deserialize(self, response, response_type):
        if response is None:
            return None
        if response_type == "file":
            return self.__deserialize_file(response)
        # fetch data from response object
        if not hasattr(response, "data"):
            response = type("obj", (object,), {"data": json.dumps(response)})
        try:
            data = json.loads(response.data)  # type: ignore
        except ValueError:
            data = response.data  # type: ignore

        if isclass(response_type) and issubclass(response_type, OpenAPISchema):
            return self.__deserialize_openapi(data, response_type)
        if get_origin(response_type) is list:
            sub_kls = get_args(response_type)[0]
            return [self.__deserialize_openapi(sub_data, sub_kls) for sub_data in data]

        return super().deserialize(response, response_type)

    def serialize(self, data):
        """
        return a sanitzed representation
        """
        return self.sanitize_for_serialization(data)

class ExtendedApi:
    """
    Extends the kubernetes.dynamic.DynamicClient to use the
    kuroboros.extended_api.ExtendedClient
    """
    _api_client: ExtendedClient
    _dynamic_client: dynamic.DynamicClient

    def __init__(self, api_client: ExtendedClient | None = None) -> None:
        if api_client is None:
            api_client = ExtendedClient()

        self._api_client = api_client
        self._dynamic_client = dynamic.DynamicClient(self._api_client)

    def _api_info_from_klass(self, klass: Type):
        av = None
        kind = None
        if isclass(klass) and issubclass(klass, CRDSchema):
            gvi = klass.get_gvi()
            assert gvi is not None
            av = f"{gvi.group}/{gvi.api_version}"
            kind = gvi.kind
        return (av, kind)

    def deserialize(self, data, klass: type):
        """
        Deserialize a OpenAPI Model into a python class
        """
        return self._api_client.deserialize(
            response=data.to_dict() if hasattr(data, "to_dict") else data,
            response_type=klass,
        )

    def serialize(self, instance):
        """
        Serialize a python class into OpenAPI Model valid value
        """
        return self._api_client.serialize(instance)

    def create(
        self,
        body: Any,
        kind: str | None = None,
        api_version: str | None = None,
        namespace: str | None = None,
        klass: Type[R] = object,
    ) -> R:
        """
        Creates the resource in the clsuter and returns it deserialized.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param body: The camelCased dictonary to create in the cluster
        :param kind: the kind to create
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param klass: the return type
        """
        av, k = (
            self._api_info_from_klass(klass)
            if (api_version, kind) == (None, None)
            else (api_version, kind)
        )
        return cast(
            R,
            self.deserialize(
                self._dynamic_client.resources.get(api_version=av, kind=k).create(
                    namespace=namespace, body=self.serialize(body)
                ),
                klass,
            ),
        )

    def get(
        self,
        api_version: str | None = None,
        kind: str | None = None,
        namespace: str | None = None,
        klass: Type[R] = object,
        **kwargs,
    ) -> R:
        """
        Gets the resources in the cluster givcen its name and returns it deserialized.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param kind: the kind to get
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param name: the target name to retrieve
        :param klass: the return type of the object

        """
        av, k = (
            self._api_info_from_klass(klass)
            if (api_version, kind) == (None, None)
            else (api_version, kind)
        )
        data = self._dynamic_client.resources.get(api_version=av, kind=k).get(
            namespace=namespace, **kwargs
        )

        ret = (
            [self.deserialize(el, get_args(klass)[0]) for el in data.items]
            if hasattr(data, "items") and not callable(data.items)
            else self.deserialize(data, klass)
        )

        return cast(R, ret)

    def patch(
        self,
        patch_body: Any,
        name: str,
        kind: str | None = None,
        api_version: str | None = None,
        namespace: str | None = None,
        subresources: list[str] | None = None,
        content_type: str = "application/merge-patch+json",
        klass: Type[R] = object,
        **kwargs,
    ) -> R:
        """
        Patch the resource in the clsuter and returns it deserialized.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param patch_body: The camelCased dictonary to create in the cluster
        :param kind: the kind to create
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param subresources: list of subresources to update
        :param content_type: the Content-Type header for the Patch api call
        :param klass: the return type
        :param **kwargs: extra arguments given to `kubernetes.dynamic.DynamicClient.patch()`

        """
        if subresources is None:
            subresources = []
        av, k = (
            self._api_info_from_klass(klass)
            if (api_version, kind) == (None, None)
            else (api_version, kind)
        )

        patch_body = self.serialize(patch_body)
        args = {
            "namespace": namespace,
            "name": name,
            "body": patch_body,
            "content_type": content_type,
        }

        resource = self._dynamic_client.resources.get(api_version=av, kind=k)
        for sub in subresources:
            if sub in resource.subresources.keys() and sub in patch_body:
                resource.subresources[sub].patch(
                    **{
                        **args,
                        **kwargs,
                        "body": {sub: patch_body.pop(sub)},
                    }
                )
        return cast(
            R,
            self.deserialize(
                resource.patch(**{**args, **kwargs}),
                klass,
            ),
        )

    def delete(
        self,
        name: str,
        api_version: str | None = None,
        kind: str | None = None,
        namespace: str | None = None,
        klass: Type = type(None),
        **kwargs,
    ):
        """
        Deletes a resource in the clsuter.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param kind: the kind to create
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param klas: the return type
        :param **kwargs: extra arguments given to `kubernetes.dynamic.DynamicClient.delete()`
        """
        av, k = (
            self._api_info_from_klass(klass)
            if (api_version, kind) == (None, None)
            else (api_version, kind)
        )
        self._dynamic_client.resources.get(api_version=av, kind=k).delete(
            namespace=namespace, name=name, **kwargs
        )



from typing import ClassVar, Dict, Tuple, TypeVar
from kubernetes import client
from kuroboros.utils import NamespaceName
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.schema.openapi_model import OpenAPISchema
from kuroboros.schema.properties import PropYaml, prop


class CRDSchema(OpenAPISchema):
    """
    Defines the CRD class for your Reconciler and Webhooks
    """

    __group_version: ClassVar[GroupVersionInfo | None]
    print_columns: ClassVar[Dict[str, Tuple[str, str]]]
    metadata: ClassVar[client.V1ObjectMeta]
    T = TypeVar("T", bound="CRDSchema")

    def __init_subclass__(cls) -> None:
        if "status" not in cls.__dict__:
            setattr(
                cls, "status", prop(dict, x_kubernetes_preserve_unknown_fields=True)
            )
        if "metadata" not in cls.__dict__:
            setattr(cls, "metadata", prop(client.V1ObjectMeta))

        if not isinstance(getattr(cls, "status"), PropYaml):
            raise RuntimeError("status must by a prop().")

        if "print_columns" not in cls.__dict__:
            cls.print_columns = {}

        return super().__init_subclass__()

    @classmethod
    def set_gvi(cls, gvi: GroupVersionInfo) -> None:
        """
        Sets the GroupVersionInfo of the class
        """
        cls.__group_version = gvi

    @classmethod
    def get_gvi(cls) -> GroupVersionInfo | None:
        """
        Gets the GroupVersionInfo of the class
        """
        return cls.__group_version

    @property
    def namespace_name(self) -> NamespaceName:
        """
        Returns a tuple of `(namespace, name)` of the resource
        """
        assert self.metadata.name is not None
        return (self.metadata.namespace, self.metadata.name)



from inspect import isclass
import json
from typing import (
    Any,
    ClassVar,
    Generic,
    Type,
    TypeVar,
    get_args,
    get_origin,
)
import threading
from datetime import timedelta
from logging import Logger
from kubernetes import client

from kuroboros.exceptions import RetriableException
from kuroboros.extended_api import ExtendedApi
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.logger import root_logger
from kuroboros.schema import OpenAPISchema, CRDSchema
from kuroboros.utils import NamespaceName, event_aware_sleep, with_timeout

T = TypeVar("T", bound=CRDSchema)
R = TypeVar("R", bound=Any)


class Result:
    """
    The result of a `reconcile` call.
    It defined if the reconcile should be requeued and in how mucho time.
    """

    requeue: bool
    requeue_after_seconds: float

    def __init__(self, requeue: bool = True, requeue_after_seconds: float = 0) -> None:
        if requeue_after_seconds < 0:
            raise RuntimeError("cannot requeue in time < 0")
        self.requeue = requeue
        self.requeue_after_seconds = requeue_after_seconds


class BaseReconciler(Generic[T]):
    """
    The base Reconciler.
    This class perform the reconcilation logic in `reconcile`
    """

    __group_version_info: ClassVar[GroupVersionInfo]
    _stop: threading.Event
    _running: bool
    _loop_thread: threading.Thread
    _namespace_name: NamespaceName
    _api_client: client.ApiClient

    reconcile_timeout: ClassVar[timedelta | None] = None
    timeout_retry: ClassVar[bool] = False
    timeout_requeue_seconds: ClassVar[float] = timedelta(minutes=5).total_seconds()

    logger: Logger
    crd_inst: T
    name: str
    extended_api: ExtendedApi

    @classmethod
    def crd_type(cls) -> Type[T]:
        """
        Return the class of the CRD
        """
        t_type = None
        for base in getattr(cls, "__orig_bases__", []):
            origin = get_origin(base)
            if origin is BaseReconciler:
                t_type = get_args(base)[0]
                break
        if t_type is None or CRDSchema not in t_type.__mro__:
            raise RuntimeError(
                "Could not determine generic type T. "
                "Subclass BaseReconciler with a concrete CRD type"
            )

        return t_type

    @classmethod
    def set_gvi(cls, gvi: GroupVersionInfo) -> None:
        """
        Sets the GroupVersionInfo of the Reconciler
        """
        cls.__group_version_info = gvi

    def __init__(self, namespace_name: NamespaceName):
        self._api_client = client.ApiClient()
        self._stop = threading.Event()
        self._running = False
        pretty_version = self.__group_version_info.pversion()
        self._namespace_name = namespace_name
        self.name = f"{self.__class__.__name__}{pretty_version}"
        self.logger = root_logger.getChild(__name__).getChild(self.__repr__())

    def __repr__(self) -> str:
        if self._namespace_name is not None:
            ns, n = self._namespace_name
            return f"{self.name}(Namespace={ns}, Name={n})"
        return f"{self.name}"

    def _deserialize(self, obj, typ):
        if isclass(typ) and issubclass(typ, CRDSchema):
            return typ(data=obj.to_dict())
        return self._api_client.deserialize(
            response=type("obj", (object,), {"data": json.dumps(obj.to_dict())}),
            response_type=typ,
        )

    def _deserialize_openapi(self, obj, typ):
        if isclass(typ) and issubclass(typ, OpenAPISchema):
            return typ(**obj)
        return self._api_client.deserialize(
            response=type("obj", (object,), {"data": json.dumps(obj)}),
            response_type=typ,
        )

    def _api_info_from_class(self, typ: Type):
        api_version = None
        kind = None
        if isclass(typ) and issubclass(typ, CRDSchema):
            gvi = typ.get_gvi()
            assert gvi is not None
            api_version = f"{gvi.group}/{gvi.api_version}"
            kind = gvi.kind

        return (api_version, kind)

    def _handle_reconcilation_loop_exception(self, e: Exception) -> Result | None:
        result = None
        if isinstance(e, client.ApiException):
            if e.status == 404:
                self.logger.info(e)
                self.logger.info(
                    "%s no longer found, killing thread", self._namespace_name
                )
            else:
                self.logger.fatal(
                    "A `APIException` ocurred while proccessing %s: %s",
                    self._namespace_name,
                    e,
                    exc_info=True,
                )
        elif isinstance(e, RetriableException):
            self.logger.warning(
                "A `RetriableException` ocurred while proccessing %s: %s",
                self._namespace_name,
                e,
            )
            result = Result(requeue_after_seconds=e.backoff)
        elif isinstance(e, TimeoutError):
            self.logger.warning(
                "A `TimeoutError` ocurred while proccessing %s: %s",
                self._namespace_name,
                e,
            )
            if not self.timeout_retry:
                self.logger.warning(
                    "`TimeoutError` will not be retried. To retry, enable it in %s",
                    self.__class__.__name__,
                )
            else:
                result = Result(requeue_after_seconds=self.timeout_requeue_seconds)
        else:
            self.logger.error(
                f"`{e.__class__.__name__}` ocurred while proccessing %s: %s",
                self._namespace_name,
                e,
                exc_info=True,
            )
        return result

    def reconcilation_loop(self):
        """
        Runs the reconciliation loop of every object
        while its a member of the `Controller`
        """
        while not self._stop.is_set():
            try:
                latest = self.get(
                    namespace=self._namespace_name[0],
                    name=self._namespace_name[1],
                    api_version=self.__group_version_info.api_version,
                    kind=self.__group_version_info.kind,
                    klass=self.crd_type(),
                )
                if self.reconcile_timeout is None:
                    result = self.reconcile(obj=latest, stopped=self._stop)
                else:
                    result = with_timeout(
                        self._stop,
                        self.timeout_retry,
                        self.reconcile_timeout.total_seconds(),
                        self.reconcile,
                        obj=latest,
                        stopped=self._stop,
                    )

            except (  # pylint: disable=broad-exception-caught
                client.ApiException,
                RetriableException,
                TimeoutError,
                Exception,
            ) as e:
                result = self._handle_reconcilation_loop_exception(e)

            should_requeue = result is not None and result.requeue
            requeue_after = result.requeue_after_seconds if result else 0

            result = None
            latest = None
            if should_requeue:
                if requeue_after == 0:
                    continue
                event_aware_sleep(self._stop, requeue_after)
            else:
                break
        self.logger.debug("%s reconcile loop stopped", self._namespace_name)

    def reconcile(
        self,
        obj: T,  # pylint: disable=unused-argument
        stopped: threading.Event,  # pylint: disable=unused-argument
    ) -> Result:  # pylint: disable=unused-argument
        """
        The function that reconcile the object to the desired status.

        :param obj: The CRD instance at the run moment
        :param stopped: The reconciliation loop event that signal a stop
        :returns result (`reconciler.Result`): Reconcilation result.
        If its `Result(requeue=False)` it wait until further updates or a
        controller restart to continue running
        """
        return Result(requeue=False)

    def get(
        self,
        api_version: str | None = None,
        kind: str | None = None,
        namespace: str | None = None,
        klass: Type[R] = object,
        **kwargs,
    ) -> R:
        """
        Gets the resources in the cluster givcen its name and returns it deserialized.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param kind: the kind to get
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param name: the target name to retrieve
        :param klass: the return type of the object

        """
        return self.extended_api.get(
            **kwargs,
            namespace=namespace,
            api_version=api_version,
            kind=kind,
            klass=klass,
        )

    def create(
        self,
        body: Any,
        kind: str | None = None,
        api_version: str | None = None,
        namespace: str | None = None,
        klass: Type[R] = object,
    ) -> R:
        """
        Creates the resource in the clsuter and returns it deserialized.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param body: The camelCased dictonary to create in the cluster
        :param kind: the kind to create
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param klass: the return type
        """
        return self.extended_api.create(
            body=body,
            kind=kind,
            api_version=api_version,
            namespace=namespace,
            klass=klass,
        )

    def patch(
        self,
        patch_body: Any,
        name: str,
        kind: str | None = None,
        api_version: str | None = None,
        namespace: str | None = None,
        subresources: list[str] | None = None,
        content_type: str = "application/merge-patch+json",
        klass: Type[R] = object,
        **kwargs,
    ) -> R:
        """
        Patch the resource in the clsuter and returns it deserialized.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param patch_body: The camelCased dictonary to create in the cluster
        :param kind: the kind to create
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param subresources: list of subresources to update
        :param content_type: the Content-Type header for the Patch api call
        :param typ: the return type
        :param **kwargs: extra arguments given to `kubernetes.dynamic.DynamicClient.patch()`

        """
        return self.extended_api.patch(
            **kwargs,
            name=name,
            api_version=api_version,
            kind=kind,
            namespace=namespace,
            klass=klass,
            subresources=subresources,
            content_type=content_type,
            patch_body=patch_body,
        )

    def delete(
        self,
        name: str,
        api_version: str | None = None,
        kind: str | None = None,
        namespace: str | None = None,
        klass: Type = type(None),
        **kwargs,
    ):
        """
        Deletes a resource in the clsuter.
        `api_version` and `kind` can be obtained from `CRDSchema` subclasses

        :param kind: the kind to create
        :param api_version: the target group/version if `None` will use prefered kind version
        :param namespace: the target namespace
        :param typ: the return type
        :param **kwargs: extra arguments given to `kubernetes.dynamic.DynamicClient.delete()`
        """
        return self.extended_api.delete(
            **kwargs,
            name=name,
            api_version=api_version,
            kind=kind,
            namespace=namespace,
            klass=klass,
        )

    def start(self, extended_api: ExtendedApi):
        """
        Starts the reconcilation loop
        """
        if self._running:
            raise RuntimeError(
                "cannot start an already started reconciler",
                f"{self.crd_type().__class__}-{self._namespace_name}",
            )
        self.extended_api = extended_api
        loop_thread = threading.Thread(
            target=self.reconcilation_loop,
            daemon=True,
            name=f"{self.name}::{self._namespace_name}",
        )
        loop_thread.start()
        self._running = True
        self._loop_thread = loop_thread

    def stop(self):
        """
        Stops the reconciliation loop
        """
        self.logger.debug("stopping %s thread", self._loop_thread.name)
        if not self.is_running():
            return
        self._stop.set()
        self._running = False
        self._api_client.close()
        del self.extended_api

    def is_running(self) -> bool:
        """
        Checks if the reconciler is running
        """
        return self._running and self._loop_thread.is_alive()



class RetriableException(Exception):
    """
    An retriable exception that will cause the reonciliation loop
    to try agains in a defined backoof
    """
    backoff: float

    def __init__(self, backoff: float, *args: object) -> None:
        self.backoff = backoff
        super().__init__(*args)


class ValidationWebhookError(Exception):
    """Exception raised for validation errors in CRD objects."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.reason = message

    def __str__(self) -> str:
        return f"ValidationWebhookError: {self.reason}"


class MutationWebhookError(Exception):
    """Exception raised for mutation errors in CRD objects."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.reason = message

    def __str__(self) -> str:
        return f"MutationWebhookError: {self.reason}"


class MultipleDefinitionsException(Exception):
    """
    Multiples types of a resource is defined under the same controller
    """

    def __init__(self, cls, ctrl, vrsn) -> None:
        super().__init__(
            f"Multiple {cls.__class__.__name__} classes found in {ctrl} {vrsn}. "
            f"Only one {cls.__class__.__name__} class is allowed per version."
        )
        self.reason = f"Multiple {cls} classes found in {ctrl} {vrsn}."

    def __str__(self) -> str:
        return f"MultipleDefinitionsException: {self.reason}"



import os
from pathlib import Path
import subprocess

from kuroboros.cli.utils import run_command_stream_simple


def kubectl_kustomize_apply(overlay: str):
    """
    execute $(which kubectl) kustomize {overlay} | $(which kubectl) apply -f
    """
    overlay = f"config/overlays/{overlay}"
    if os.path.exists(os.path.join(Path().absolute(), overlay)):
        kubectl = subprocess.run(
            "which kubectl",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
        if kubectl == "":
            raise RuntimeError(
                "kubectl not found, please install it in your $PATH first"
            )
        kubectl_cmd = kubectl.stdout.strip()
        run_command_stream_simple(
            f"{kubectl_cmd} kustomize {overlay} | {kubectl_cmd} apply -f-"
        )
    else:
        raise FileNotFoundError(f"{overlay} does not exists")



from kuroboros.config import KuroborosConfig
from kuroboros.cli.utils import get_image_info, run_command_stream_simple


def build_operator_image() -> None:
    """
    Builds a Docker image with given build-args and image tag
    """
    binary = KuroborosConfig.get("build", "builder", "binary", typ=str)
    args = KuroborosConfig.get("build", "builder", "args", typ=list)
    img = get_image_info()
    command = " ".join([binary, " ".join(args)])
    run_command_stream_simple(command, env={"IMG": img})


from typing import Dict, List, cast

from kuroboros.config import KuroborosConfig
from kuroboros.controller import ControllerConfig
from kuroboros.group_version_info import GroupVersionInfo
from kuroboros.schema import CRDSchema, PropYaml
from kuroboros.cli.utils import x_kubernetes_kebab, yaml_format
import kuroboros.cli.templates as temps

temps.env.filters["maybekebab"] = x_kubernetes_kebab
temps.env.filters["yaml"] = yaml_format


def crd_schema(
    versions: Dict[str, CRDSchema], group_version_info: GroupVersionInfo
) -> str:
    """
    Generates the `CustomResourceDefinition` for the inherited `CRDSchema` class
    """
    version_print_columns = {}
    version_props = {}
    version_desc = {}
    for version in versions:
        crd = versions[version]
        props = {}

        base_attr = dir(CRDSchema)
        child_attr = [attr for attr in dir(crd) if attr not in base_attr]

        for attr_name in child_attr:
            if attr_name in base_attr:
                continue
            attr = object.__getattribute__(crd, attr_name)
            if isinstance(attr, PropYaml):
                cased_attr_name = crd.attribute_map[attr_name]
                props[cased_attr_name] = attr

        if cast(PropYaml, crd.status).typ != "object":
            raise TypeError("status can only be a `dict` type object")

        print_columns: List[Dict] = []
        for name, (path, typ) in crd.print_columns.items():
            print_columns.append({"json_path": path, "name": name, "type": typ})
        if len(print_columns) > 0:
            version_print_columns[version] = print_columns
        if crd.__doc__ is not None:
            version_desc[version] = crd.__doc__.strip()
        version_props[version] = {"props": props}

    crd_template = temps.env.get_template("generate/crd/crd.yaml.j2")
    return crd_template.render(
        gvi=group_version_info,
        version_props=version_props,
        version_desc=version_desc,
        version_print_columns=version_print_columns,
    )


def rbac_sa() -> str:
    """
    Generates the operator `ServiceAccount`
    """
    rbac_sa_template = temps.env.get_template("generate/rbac/service-account.yaml.j2")
    return rbac_sa_template.render(name=KuroborosConfig.get("operator", "name"))


def rbac_operator_role(controllers: List[ControllerConfig]) -> str:
    """
    Generates the operator `Role`.
    Loads custom `Policies` to use in the `Role` from all
    the sections that start with `generate.rbac.policies.`
    """
    policies_conf = KuroborosConfig.get("generate", "rbac", "policies")

    policies = []
    for policy in cast(list[dict], policies_conf):
        policy_obj = {
            "api_groups": policy["api_groups"] or [],
            "resources": policy["resources"] or [],
            "verbs": policy["verbs"] or [],
        }
        policies.append(policy_obj)

    for ctrl in controllers:
        ctrl_crd_policy = {
            "api_groups": [ctrl.group_version_info.group],
            "resources": [
                ctrl.group_version_info.plural,
                f"{ctrl.group_version_info.plural}/status",
            ],
            "verbs": ["create", "list", "watch", "delete", "get", "patch", "update"],
        }
        policies.append(ctrl_crd_policy)

    rbac_operator_role_template = temps.env.get_template(
        "generate/rbac/operator-role.yaml.j2"
    )
    return rbac_operator_role_template.render(
        name=KuroborosConfig.get("operator", "name"), policies=policies
    )


def rbac_leader_role() -> str:
    """
    Generates leader election `Role`
    """
    rbac_leader_election_role_template = temps.env.get_template(
        "generate/rbac/leader-election-role.yaml.j2"
    )
    return rbac_leader_election_role_template.render(
        name=KuroborosConfig.get("operator", "name")
    )


def rbac_operator_role_binding() -> str:
    """
    Generates the operator `RoleBinding` of the `ServiceAccount` and the `Role`
    """
    rbac_operator_role_binding_template = temps.env.get_template(
        "generate/rbac/operator-role-binding.yaml.j2"
    )
    return rbac_operator_role_binding_template.render(
        name=KuroborosConfig.get("operator", "name")
    )


def rbac_leader_role_binding() -> str:
    """
    Generates the leader election `RoleBinding` of the `ServiceAccount` and the `Role`
    """
    rbac_leader_election_role_binding_template = temps.env.get_template(
        "generate/rbac/leader-election-role-binding.yaml.j2"
    )
    return rbac_leader_election_role_binding_template.render(
        name=KuroborosConfig.get("operator", "name")
    )


def operator_deployment() -> str:
    """
    Generates the `Deployment` of the operator.
    takes the `image` spec of the container from the `generate.deployment.image` section
    of the `config_file` passed in the arguments
    """
    deployment_template = temps.env.get_template(
        "generate/deployment/operator-deployment.yaml.j2"
    )
    return deployment_template.render(name=KuroborosConfig.get("operator", "name"))


def operator_metrics_service() -> str:
    """
    Generates the `Service` for the operator metrics.
    The service is used to expose the operator's metrics server
    """
    deployment_metrics_service = temps.env.get_template(
        "generate/deployment/metrics-service.yaml.j2"
    )
    return deployment_metrics_service.render(
        name=KuroborosConfig.get("operator", "name")
    )


def operator_webhook_service() -> str:
    """
    Generates the `Service` for the operator webhook.
    The service is used to expose the operator's webhook server
    """
    deployment_webhook_service = temps.env.get_template(
        "generate/deployment/webhook-service.yaml.j2"
    )
    return deployment_webhook_service.render(
        name=KuroborosConfig.get("operator", "name")
    )


def operator_config() -> str:
    """
    Generates the `ConfigMap` from the `config_file` passed in the parameters
    for the `Deployment` to use. Only takes the `operator` section of the file
    """
    kuroboros_config = KuroborosConfig.dumps("operator")
    deployment_config_template = temps.env.get_template(
        "generate/deployment/operator-config.yaml.j2"
    )
    return deployment_config_template.render(
        name=KuroborosConfig.get("operator", "name"), config=kuroboros_config
    )


def validation_webhook_config(controllers: List[ControllerConfig]) -> str:
    """
    Generates the `ValidatingWebhookConfiguration` for the controllers
    """
    validation_webhooks: List[Dict] = []
    for ctrl in controllers:
        if ctrl.validation_webhook is not None:
            validation_webhooks.append(ctrl.validation_webhook.get_config_dict())

    validation_webhook_configs = temps.env.get_template(
        "generate/webhooks/validation-webhook-config.yaml.j2"
    )
    return validation_webhook_configs.render(
        name=KuroborosConfig.get("operator", "name"), webhooks=validation_webhooks
    )


def mutation_webhook_config(controllers: List[ControllerConfig]) -> str:
    """
    Generates the `ValidatingWebhookConfiguration` for the controllers
    """
    mutation_webhooks: List[Dict] = []
    for ctrl in controllers:
        if ctrl.mutation_webhook is not None:
            mutation_webhooks.append(ctrl.mutation_webhook.get_config_dict())

    mutation_webhook_configs = temps.env.get_template(
        "generate/webhooks/mutation-webhook-config.yaml.j2"
    )
    return mutation_webhook_configs.render(
        name=KuroborosConfig.get("operator", "name"), webhooks=mutation_webhooks
    )


def kustomize_file(resources: list[str], images: list[dict] | None = None):
    """
    Generates a Kustomization file
    """
    if images is None:
        images = []

    kustomization_template = temps.env.get_template("generate/kustomization.yaml.j2")
    return kustomization_template.render(resources=resources, images=images)



from typing import Any, Dict, Type, cast, TypeVar

import tomlkit

OPERATOR_NAMESPACE = "default"
try:
    with open(
        "/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r", encoding="utf-8"
    ) as f:
        ns = f.read().strip()
        OPERATOR_NAMESPACE = ns
except Exception:  # pylint: disable=broad-exception-caught
    pass


T = TypeVar("T")


class KuroborosConfig:
    """
    Configuration class
    """

    _config: Dict[str, Any] = {
        "operator": {
            "name": "kuroboros-operator",
            "leader_acquire_interval_seconds": 10.0,
            "log_level": "INFO",
            "cleanup_interval_seconds": 5.0,
            "metrics": {
                "interval_seconds": 5.0,
                "port": 8080,
            },
            "webhook_server": {
                "port": 443,
                "cert_path": "/etc/tls/tls.crt",
                "key_path": "/etc/tls/tls.key",
                "gunicorn_workers": 1,
            },
        },
        "build": {
            "builder": {
                "binary": "docker",
                "args": ["build", ".", "-t", "$IMG"],
            },
            "image": {
                "registry": "",
                "repository": "kuroboros-operator",
                "tag_prefix": "",
                "tag": "$PYPROJECT_VERSION",
                "tag_suffix": "",
            },
        },
        "generate": {"rbac": {"policies": []}},
    }

    @classmethod
    def get(cls, *keys: *tuple[str, ...], typ: Type[T] = cast(Type[T], None)) -> T:
        """
        Gets a config field from the toml field and optionally casts it to the given type.
        """
        current = cls._config
        for key in keys:
            if key in current:
                current = current[key]
            else:
                raise KeyError(f"config for {'.'.join(keys)} not found")
        if typ is not None:
            if typ == float and isinstance(current, int):
                return int(current) # type: ignore
            if not isinstance(current, typ):
                raise AssertionError(
                    f"type of {'.'.join(keys)} ({current.__class__}) is not {typ}"
                )
            return current.unwrap() if hasattr(current, "unwrap") else current # type: ignore
        return current.unwrap() if hasattr(current, "unwrap") else current

    @staticmethod
    def _merge(default, user):
        for key, value in user.items():
            if (
                isinstance(value, dict)
                and key in default
                and isinstance(default[key], dict)
            ):
                default[key] = KuroborosConfig._merge(default[key], value)
            else:
                default[key] = value
        return default

    @classmethod
    def load(cls, path: str):
        """
        Loads a TOML file as the config for the operator
        """
        toml_config = None
        try:
            with open(path, "r", encoding="utf-8") as config_file:
                content = config_file.read()
                toml_config = tomlkit.loads(content)

            if toml_config is not None:
                cls._config = cls._merge(cls._config, toml_config)
        except FileNotFoundError:
            pass

    @classmethod
    def dumps(cls, key: str) -> str:
        """
        Dumps the given section into a string
        """
        temp = tomlkit.document()
        temp[key] = cls._config[key]

        return tomlkit.dumps(temp)



import concurrent
import concurrent.futures
import threading
import time
from typing import Callable, ParamSpec, Tuple, TypeVar


NamespaceName = Tuple[str | None, str]


def event_aware_sleep(event: threading.Event, timeout: float):
    """
    Checks for `event.is_set()` every 0.1 seconds within the `timeout`
    range. If the `event` is set in any time, the sleep will finish and return
    """
    start_time = time.time()
    remaining = timeout
    while remaining > 0:
        if event.is_set():
            return
        sleep_time = min(remaining, 0.1)
        time.sleep(sleep_time)
        elapsed = time.time() - start_time
        remaining = timeout - elapsed
    return


T = TypeVar("T")
P = ParamSpec("P")


def with_timeout(
    stop: threading.Event,
    retriable: bool,
    timeout_seconds: float | None,
    func: Callable[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    """
    Runs a function in a ThreadPoolExecutor and send a cancel event to it on SIGINT or timeout
    """
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
    future = executor.submit(func, *args, **kwargs)
    start_time = time.time()

    try:
        while not stop.is_set():
            # Calculate remaining time if timeout is specified
            current_time = time.time()
            if timeout_seconds is not None:
                elapsed = current_time - start_time
                if elapsed >= timeout_seconds:
                    if not retriable:
                        stop.set()
                    raise TimeoutError(
                        f"`{func.__name__}` timed out after {timeout_seconds} seconds"
                    )
                wait_time = min(0.1, timeout_seconds - elapsed)
            else:
                wait_time = 0.1

            # Check future completion with dynamic timeout
            try:
                return future.result(timeout=wait_time)
            except concurrent.futures.TimeoutError:
                # Continue loop to check stop event or timeout
                pass

        # Stop event was set
        raise InterruptedError(f"`{func.__name__}` was stopped before completion")
    finally:
        # Cancel future if not completed
        if not future.done():
            future.cancel()
        # Shutdown executor without waiting for thread termination
        executor.shutdown(wait=False)



import re
from typing import cast

import inflect

from kuroboros.utils import NamespaceName


class GroupVersionInfo:
    """
    Contains the static information about the version to run in the controller
    """

    __STABILITY_ORDER = {"alpha": 0, "beta": 1, "stable": 2}
    __VERSION_PATTERN = r"^v(\d+)(?:([a-z]+)(\d+))?$"

    api_version: str
    group: str
    major: int
    stability: str
    minor: int

    kind: str
    singular: str
    plural: str
    crd_name: str
    short_names: list[str]
    scope: str

    @staticmethod
    def is_valid_api_version(api_version: str) -> bool:
        """
        Validates if the given API version string matches the expected format.
        """
        return re.match(GroupVersionInfo.__VERSION_PATTERN, api_version) is not None

    def is_namespaced(self):
        """
        Returns `True` if the GVI is scoped `Namespaced`
        """
        return self.scope == "Namespaced"

    def pkind(self, namespace_name: NamespaceName | None = None) -> str:
        """
        Return a string to represent the CRD as `MyCRDV1Stable` or
        `MyCRD(Namespace="string", Name="string")` if available
        """
        if namespace_name is not None and namespace_name != (None, None):
            ns = namespace_name[0]
            n = namespace_name[1]
            return f"{self.kind}{self.pversion()}(Namespace={ns}, Name={n})"
        return f"{self.kind}{self.pversion()}"

    def pversion(self) -> str:
        """
        Get a pretty version string as `V1Stable`
        """
        major = self.major
        stability = self.stability.capitalize()
        minor = self.minor if self.minor != 0 else ""

        return f"V{major}{stability}{minor}"

    def __init__(
        self,
        api_version: str,
        group: str,
        kind: str,
        scope: str = "Namespaced",
        **kwargs,
    ):
        inf = inflect.engine()
        self.api_version = api_version
        self.group = group
        self.scope = scope

        if self.scope not in ("Namespaced", "Cluster"):
            raise ValueError("scope must be one of `Namespaced` or `CLuster`")

        match = re.match(self.__VERSION_PATTERN, self.api_version)
        if not match:
            raise ValueError(f"Invalid format {self.api_version}")

        self.major = int(match.group(1))
        stability = match.group(2) or "stable"
        self.minor = int(match.group(3)) if match.group(3) else 0

        if stability not in self.__STABILITY_ORDER:
            raise ValueError(f"Unknown stability level: {stability}")
        self.stability = stability

        self.kind = kind
        self.singular = kwargs.get("singular", kind.lower())
        self.plural = kwargs.get(
            "plural", inf.plural_noun(cast(inflect.Word, kind)).lower()
        )
        self.crd_name = kwargs.get("crd_name", f"{self.plural}.{self.group}")
        self.short_names = kwargs.get("short_names", [])

    def _key(self):
        return (self.major, self.__STABILITY_ORDER[self.stability], self.minor)

    def __eq__(self, other):
        return self._key() == other._key()

    def __lt__(self, other):
        return self._key() < other._key()

    def __repr__(self):
        return f"GroupVersionInfo(version={self.api_version}, group={self.group})"



import os
from pathlib import Path
from importlib.metadata import version as pyver
import sys

import click

from kuroboros.cli.build import build_operator_image
from kuroboros.cli.generate import (
    crd_schema,
    kustomize_file,
    mutation_webhook_config,
    operator_config,
    operator_deployment,
    operator_metrics_service,
    operator_webhook_service,
    rbac_leader_role,
    rbac_leader_role_binding,
    rbac_operator_role,
    rbac_operator_role_binding,
    rbac_sa,
    validation_webhook_config,
)

from kuroboros.cli.new import (
    new_config,
    new_crd,
    new_dockerfile,
    new_group_versions,
    new_mutation_webhook,
    new_pyproject,
    new_reconciler,
    new_validation_webhook,
)
from kuroboros.cli.deploy import kubectl_kustomize_apply
from kuroboros.cli.utils import create_file, get_image_info, load_controller_configs
from kuroboros.config import KuroborosConfig
from kuroboros.operator import Operator
from kuroboros.logger import root_logger


VERSION_NUM = pyver("kuroboros")


KUSTOMIZE_OUT = "config/base"
KUSTOMIZE_OVERLAYS = "config/overlays"
CRD_OUT = "crd"
RBAC_OUT = "rbac"
WEBHOOKS_OUT = "webhooks"
DEPLOYMENT_OUT = "deployment"
CONTROLLERS_PATH = "controllers"

sys.path.insert(0, str(Path().absolute()))


@click.group(help=f"Kuroboros Framework {VERSION_NUM}")
@click.option(
    "-c",
    "--config",
    "config_file",
    default="operator.toml",
    help="Configuration file to use [default: opearator.toml]",
)
@click.option(
    "-l",
    "--log-level",
    "log_level",
    help="Log level for the app",
)
@click.pass_context
def cli(ctx: click.Context, config_file, log_level):
    """
    The CLI entrypoint

    Kuroboros Framework {VERSION_NUM}
    """

    ctx.ensure_object(dict)
    ctx.obj["config_file"] = config_file
    KuroborosConfig.load(config_file)
    if log_level is not None:
        root_logger.setLevel(log_level.upper())
    else:
        root_logger.setLevel(KuroborosConfig.get("operator", "log_level", typ=str))
    ctx.obj["controllers"] = load_controller_configs(CONTROLLERS_PATH)


@cli.command("version", help="Get kuroboros version")
def version_cli():
    """
    Get kuroboros version
    """
    click.echo(VERSION_NUM)


@cli.group(help="Generate the kubernetes resources manifests to deploy the operator")
@click.pass_context
def generate(ctx: click.Context):  # pylint: disable=unused-argument
    """
    Generate the kubernetes resources manifests to deploy the operator
    """


@generate.command(help="Generates the CRDs YAML manifests")
@click.pass_context
def crd(ctx):
    """
    Generates the CRDs YAML manifests
    """
    controllers = ctx.obj["controllers"]
    click.echo("🌀 Generating CRD YAMLs")
    click.echo(f"{KUSTOMIZE_OUT}/{CRD_OUT}/")
    output = os.path.join(Path().absolute(), KUSTOMIZE_OUT, CRD_OUT)

    resources = []
    versions_dict = {}
    for ctrl_conf in controllers:
        versions_dict[ctrl_conf.name] = {}
        for version in ctrl_conf.versions:
            versions_dict[ctrl_conf.name][version.name] = version.crd

        create_file(
            output,
            f"{ctrl_conf.group_version_info.kind}.yaml",
            crd_schema(versions_dict[ctrl_conf.name], ctrl_conf.group_version_info),
        )
        resources.append(f"{ctrl_conf.group_version_info.kind}.yaml")

    create_file(output, "kustomization.yaml", kustomize_file(resources))


@generate.command(help="Generates the RBAC YAML manifests")
@click.pass_context
def rbac(ctx):
    """
    Generates the RBAC YAML manifests
    """
    controllers = ctx.obj["controllers"]
    click.echo("🌀 Generating RBAC YAMLs")
    click.echo(f"{KUSTOMIZE_OUT}/{RBAC_OUT}/")
    output = os.path.join(Path().absolute(), KUSTOMIZE_OUT, RBAC_OUT)

    resources = [
        "service-account.yaml",
        "operator-role.yaml",
        "operator-role-binding.yaml",
        "leader-election-role.yaml",
        "leader-election-role-binding.yaml",
    ]

    create_file(output, "service-account.yaml", rbac_sa())
    create_file(output, "operator-role.yaml", rbac_operator_role(controllers))
    create_file(output, "operator-role-binding.yaml", rbac_operator_role_binding())
    create_file(output, "leader-election-role.yaml", rbac_leader_role())
    create_file(output, "leader-election-role-binding.yaml", rbac_leader_role_binding())
    create_file(output, "kustomization.yaml", kustomize_file(resources))


@generate.command(help="Generates the Webhooks YAML manifests")
@click.pass_context
def webhooks(ctx):
    """
    Generates the Webhooks YAML manifests
    """
    controllers = ctx.obj["controllers"]
    click.echo("🌀 Generating Webhooks YAMLs")
    click.echo(f"{KUSTOMIZE_OUT}/{WEBHOOKS_OUT}/")
    output = os.path.join(Path().absolute(), KUSTOMIZE_OUT, WEBHOOKS_OUT)

    resources = []
    ctrls_with_validation_webhooks = [
        ctrl for ctrl in controllers if ctrl.validation_webhook is not None
    ]
    ctrls_with_mutation_webhooks = [
        ctrl for ctrl in controllers if ctrl.mutation_webhook is not None
    ]

    if len(ctrls_with_validation_webhooks) > 0:
        create_file(
            output,
            "validation-webhooks.yaml",
            validation_webhook_config(ctrls_with_validation_webhooks),
        )
        resources.append("validation-webhooks.yaml")

    if len(ctrls_with_mutation_webhooks) > 0:
        create_file(
            output,
            "mutation-webhooks.yaml",
            mutation_webhook_config(ctrls_with_mutation_webhooks),
        )
        resources.append("mutation-webhooks.yaml")
    if len(resources) > 0:
        create_file(output, "kustomization.yaml", kustomize_file(resources))
    else:
        click.echo("Nothing to create")


@generate.command(help="Generates the Deployment YAML manifests")
@click.pass_context
def deployment(ctx):
    """
    Generates the Deployment YAML manifests
    """
    click.echo("🌀 Generating Deployment YAMLs")
    click.echo(f"{KUSTOMIZE_OUT}/{DEPLOYMENT_OUT}/")
    output = os.path.join(Path().absolute(), KUSTOMIZE_OUT, DEPLOYMENT_OUT)

    controllers = ctx.obj["controllers"]
    resources = [
        "operator-deployment.yaml",
        "operator-config.yaml",
        "metrics-service.yaml",
    ]
    include_webhook_service = False
    for ctrl in controllers:
        if ctrl.has_webhooks():
            include_webhook_service = True
    image_config = []

    img = get_image_info()

    if img != "kuroboros-operator:latest":
        image_config = [
            {
                "name": "kuroboros-operator:latest",
                "new_name": img.split(":")[0],
                "new_tag": img.split(":")[1],
            }
        ]

    if include_webhook_service:
        create_file(output, "webhook-service.yaml", operator_webhook_service())
        resources.append("webhook-service.yaml")
    create_file(output, "metrics-service.yaml", operator_metrics_service())
    create_file(output, "operator-deployment.yaml", operator_deployment())
    create_file(output, "operator-config.yaml", operator_config())
    create_file(output, "kustomization.yaml", kustomize_file(resources, image_config))


@generate.command(help="Generates all the YAML manifests")
@click.pass_context
def manifests(ctx):
    """
    Generates all the YAML manifests
    """
    ctx.invoke(crd)
    ctx.invoke(rbac)
    ctx.invoke(deployment)
    ctx.invoke(webhooks)


@generate.command(help="Generate a new overlay in config/overlays")
@click.argument("name")
@click.pass_context
def overlay(ctx, name):
    """
    Generate a new overlay in config/overlays
    """
    controllers = ctx.obj["controllers"]
    click.echo(f"🌀 Creating new overlay {name}")
    output = os.path.join(Path().absolute(), KUSTOMIZE_OVERLAYS, name)
    paths = ["../../base/rbac", "../../base/crd", "../../base/deployment"]
    for ctrl in controllers:
        if ctrl.has_webhooks():
            paths.append("../../base/webhooks")
            break

    file = kustomize_file(paths)
    create_file(output, "kustomization.yaml", file)


@cli.group(help="Creates a new Kuroboros Resource")
def new():
    """
    Creates a new Kuroboros Resource
    """


@new.command(help="Creates a Controller with a base version, a reconciler and its CRD")
@click.option("--kind", type=str, required=True, help="The kind of the CRD")
@click.option(
    "--api-version",
    type=str,
    required=True,
    help="The version to use (example: v1alpha1)",
)
@click.option("--group", type=str, required=True, help="The group owner of the CRD")
def controller(kind: str, api_version: str, group: str):
    """
    Creates a Controller with a base version, a reconciler and its CRD
    """
    click.echo(f"🐍 Creating {kind} Controller")
    python_module = f"{CONTROLLERS_PATH}.{kind.lower()}.{api_version}"
    rec = new_reconciler(kind, python_module)
    crd_data = new_crd(kind)
    group_versions = new_group_versions(api_version, group, kind)

    click.echo(f"{CONTROLLERS_PATH}/")
    create_file(
        f"{CONTROLLERS_PATH}/{kind.lower()}",
        "group_version.py",
        group_versions,
        overwrite=False,
    )
    create_file(
        f"{CONTROLLERS_PATH}/{kind.lower()}/{api_version}",
        "reconciler.py",
        rec,
        overwrite=False,
    )
    create_file(
        f"{CONTROLLERS_PATH}/{kind.lower()}/{api_version}",
        "crd.py",
        crd_data,
        overwrite=False,
    )


@new.command(help="Creates a Webhook for a Controller given its kind and version")
@click.option("--kind", type=str, required=True, help="The kind of the CRD")
@click.option(
    "--api-version",
    type=str,
    required=True,
    help="The version to use (example: v1alpha1)",
)
@click.option(
    "--type",
    "typ",
    type=str,
    required=True,
    help="The type of webhook to create (validation, mutation)",
)
def webhook(kind: str, api_version: str, typ: str):
    """
    Creates a Webhook for a Controller given its kind and version
    """
    if typ not in ("validation", "mutation"):
        raise ValueError("type must be one of (validation, mutation)")
    click.echo(f"🐍 Creating {kind} {typ.capitalize()}Webhook")
    python_module = f"{CONTROLLERS_PATH}.{kind.lower()}.{api_version}"
    wh = None
    if typ == "mutation":
        wh = new_mutation_webhook(kind, python_module)
    else:
        wh = new_validation_webhook(kind, python_module)

    click.echo(f"{CONTROLLERS_PATH}/")
    create_file(
        f"{CONTROLLERS_PATH}/{kind.lower()}/{api_version}",
        f"{typ}.py",
        wh,
        overwrite=False,
        parents=False,
    )


@new.command(help="Creates a new Kuroboros Operator project")
@click.argument("name", type=str)
def operator(name):
    """
    Creates a new Kuroboros Operator project
    """
    click.echo(f"🌀🐍 Creating {name} Operator")
    conf = new_config(name)
    dockerfile = new_dockerfile()
    pyproject = new_pyproject(name, VERSION_NUM)

    create_file(".", "operator.toml", conf)
    create_file(".", "pyproject.toml", pyproject)
    create_file(".", "Dockerfile", dockerfile)
    create_file("controllers", "__init__.py", "")


@cli.command(help="Applies the given overlay to the current context")
@click.argument("overlay", type=str)
def deploy(overlay):  # pylint: disable=redefined-outer-name
    """
    Applies the given overlay to the current context
    """
    click.echo(f"🌀 Deploying Operator from {overlay} overlay")
    kubectl_kustomize_apply(overlay)


@cli.command(help="Build the image")
def build():
    """
    Build the image
    """
    build_operator_image()


@cli.command(help="Starts the Kuroboros Operator")
@click.option(
    "--skip-controllers",
    is_flag=True,
    default=False,
    help="Skips all controllers startup",
)
@click.option(
    "--skip-webhook-server",
    is_flag=True,
    default=False,
    help="Skips the webhook server startup",
)
@click.pass_context
def start(ctx, skip_controllers, skip_webhook_server):
    """
    Starts the Kuroboros Operator

    --skip-controllers: Skips all controllers startup
    --skip-webhook-server: Skips the webhook server startup
    """
    controllers = ctx.obj["controllers"]
    op = Operator()
    click.echo(f"🌀🐍 Starting {op.name} ...")
    op.start(
        controllers=controllers,
        skip_webhook_server=skip_webhook_server,
        skip_controllers=skip_controllers,
    )


if __name__ == "__main__":
    cli(auto_envvar_prefix="KUROBOROS")  # pylint: disable=no-value-for-parameter



import sys
import kuroboros.cli.templates as temps


def new_crd(kind: str) -> str:
    """
    Creates the new CRD python class file
    """
    crd_template = temps.env.get_template("new/controller/types.py.j2")
    return crd_template.render(kind=kind)


def new_reconciler(kind: str, module: str) -> str:
    """
    Creates the new reconciler python class file
    """
    reconciler_template = temps.env.get_template("new/controller/reconciler.py.j2")
    return reconciler_template.render(kind=kind, module=module)


def new_config(name):
    """
    Creates a base operator.toml
    """
    conf_template = temps.env.get_template("new/project/operator.toml.j2")
    return conf_template.render(name=name)


def new_dockerfile():
    """
    Creates a base Dockerfile
    """
    version = ".".join([str(num) for num in sys.version_info[0:3]])
    dockerfile_template = temps.env.get_template("new/project/docker.j2")
    return dockerfile_template.render(python_version=version)

def new_pyproject(name, version) -> str:
    """
    Creates the pyproject.toml file of the operator
    """
    pyproject_template = temps.env.get_template("new/project/pyproject.toml.j2")
    return pyproject_template.render(name=name, kuroboros_version=version)


def new_group_versions(version: str, group: str, kind: str):
    """
    Creates the controller GVI
    """
    group_version_template = temps.env.get_template(
        "new/controller/group_version.py.j2"
    )
    return group_version_template.render(version=version, group=group, kind=kind)


def new_validation_webhook(kind: str, module: str):
    """
    Creates a base validation webhook
    """
    validation_template = temps.env.get_template("new/webhooks/validation.py.j2")
    return validation_template.render(kind=kind, module=module)


def new_mutation_webhook(kind: str, module: str):
    """
    Creates a base mutation webhook
    """
    mutation_template = temps.env.get_template("new/webhooks/mutation.py.j2")
    return mutation_template.render(kind=kind, module=module)
